/*
**    Mesh Converter
**    Copyright (C) 2009  Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public Licens1e for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MC_CONNECTIVITY_INDEX_H_
#define MC_CONNECTIVITY_INDEX_H_

#include <MC_int_vector_vector.hpp>

#include <map>
#include <set>


namespace mesh_conv
{

    class MC_int_pair;
    class MC_int_pair_less;


    /// class of indexed connectivity stored as a vector
    class MC_connectivity_index : public MC_int_vector_vector
    {
    public:

        //************************************************//
        //************************************************//
        //CONSTRUCTORS
        //************************************************//
        //************************************************//

        /** \brief standard constructor */
        MC_connectivity_index();
        /** \brief direct constructor from some MC_int_vector */
        MC_connectivity_index(const MC_int_vector& v0);
        /** \brief direct constructor from some MC_int_vector */
        MC_connectivity_index(const MC_int_vector& v0,const MC_int_vector& v1);
        /** \brief direct constructor from some MC_int_vector */
        MC_connectivity_index(const MC_int_vector& v0,const MC_int_vector& v1,const MC_int_vector& v2);
        /** \brief direct constructor from some MC_int_vector */
        MC_connectivity_index(const MC_int_vector& v0,const MC_int_vector& v1,const MC_int_vector& v2,const MC_int_vector& v3);
        /** \brief constructor from MC_int_vector_vector */
        MC_connectivity_index(const MC_int_vector_vector& v0);
        /** \brief copy constructor */
        MC_connectivity_index(const MC_connectivity_index& v0);



        //************************************************//
        //************************************************//
        //Neighbooring
        //************************************************//
        //************************************************//

        /** \brief build the one ring of the connectivity */
        MC_connectivity_index& build_ring();
        /** \brief return the one ring */
        const std::map<int,std::set<int> >& ring() const;

        /** \brief build the one star of the connectivity */
        MC_connectivity_index& build_star();
        /** \brief return the one star (polygon neighbors of a vertex) */
        const std::map<int,std::set<int> >& star() const;

        /** \brief get the polygon neighboring of an other polygon */
        std::map<int,std::set<int> > polygon_neighbors() const;

        /** \brief rebuild the index */
        MC_connectivity_index& update_neighbors();


        //************************************************//
        //************************************************//
        //informations
        //************************************************//
        //************************************************//

        /** \brief get the used vertices
            \warning need to compute a set on every polygons
        */
        std::set<int> used_vertex() const;

        /** \brief is the connectivity only made of triangles */
        bool is_triangle_only() const;


        //************************************************//
        //************************************************//
        //Modification
        //************************************************//
        //************************************************//

        /** \brief triangulate the connectivity

            Do fans starting at the first vertex of the polygon
        */
        MC_connectivity_index triangulated() const;

        /** \brief flip every polygon of the mesh */
        MC_connectivity_index fliped_polygon() const;

        //************************************************//
        //************************************************//
        //Boundary
        //************************************************//
        //************************************************//

        /** \brief get the edges on the boundary */
        std::set <MC_int_pair,MC_int_pair_less> boundary() const;
        /** \brief get the index of the vertices on the boundary */
        std::set<int> boundary_vertex() const;
        /** \brief get the index of the polygons having at least one vertex on the boundary

            \note assume one star is computed
        */
        std::set<int> boundary_polygon() const;


        //************************************************//
        //************************************************//
        //Edges
        //************************************************//
        //************************************************//

        /** \brief get every edge of the mesh */
        std::set<MC_int_pair,MC_int_pair_less> edges() const;

        //************************************************//
        //************************************************//
        //Connectivity algorithm
        //************************************************//
        //************************************************//

        /** \brief Perform distance propagation on triangles using a per_vertex field */
        MC_int_vector triangle_propagation_using_vertex_field(const MC_int_vector& seed_vertices,const MC_double_vector& field_per_vertex,const double& threshold,const std::set<int>& invalidated_triangle) const;



    private:

        /** one ring of the connectivity */
        std::map<int,std::set<int> > ring_1;
        /** one star of vertices */
        std::map<int,std::set<int> > star_1;
    };
}

#endif
