/*
**    Mesh Converter
**    Copyright (C) 2009  Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public Licens1e for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef _MC_CURVE_HPP_
#define _MC_CURVE_HPP_

#include <MC_v3d_vector.hpp>
#include <MC_int_vector.hpp>

#include <list>
#include <map>

#include <stdlib.h>

namespace mesh_conv
{
    class MC_segment;
    class MC_int_pair_unique;

    /** \brief class of 3D curve extending the MC_v3d_vector class. Ordering is important.

      Vertices are stored in a vertex internally
    */
    class MC_curve : public MC_v3d_vector
    {
        public:

        //************************************************//
        //************************************************//
        //CONSTRUCTORS
        //************************************************//
        //************************************************//

        /** \brief empty constructor */
        MC_curve();
        /** \brief direct constructor with a MC_v3d */
        MC_curve(const MC_v3d& v0);
        /** \brief direct constructor with a MC_v3d */
        MC_curve(const MC_v3d& v0,const MC_v3d& v1);
        /** \brief direct constructor with a MC_v3d */
        MC_curve(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2);
        /** \brief direct constructor with a MC_v3d */
        MC_curve(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2,const MC_v3d& v3);
        /** \brief direct constructor with a MC_v3d */
        MC_curve(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2,const MC_v3d& v3,const MC_v3d& v4);
        /** \brief direct constructor from a MC_v3d_vector */
        MC_curve(const MC_v3d_vector& vec);
        /** \brief copy constructor */
        MC_curve(const MC_curve& vec);
        /** \brief direct constructor from a vector of x,y and z coordinates */
        MC_curve(const MC_double_vector& x_vector,const MC_double_vector& y_vector,const MC_double_vector& z_vector);
        /** \brief direct constructor from a segment */
        MC_curve(const MC_segment& seg);

        /** \brief direct constructor from a std::list */
        MC_curve(const std::list<MC_v3d>& input);

        //************************************************//
        //************************************************//
        //Information
        //************************************************//
        //************************************************//

        /** \brief return the length of the curve */
        double length() const;

        /** \brief get the closest point on the curve
            \return MC_v3d the closest point on the curve
            \return int the segment number corresponding
            \return double the relative position on this segment
        */
        std::pair <MC_v3d,std::pair<int,double> > closest_point(const MC_v3d& x) const;

        /** \brief return barycenter of a curve

            This is not the same as barycenter of a MC_v3d_vector, the weighting of the curve length is considered
        */
        MC_v3d barycenter() const;

        //************************************************//
        //************************************************//
        //Manipulation
        //************************************************//
        //************************************************//

        /** \brief unclose a mesh if it was closed (ensure c[end]!=c[0])
            \return <the unclosed mesh,was it closed ?>
        */
        std::pair <MC_curve,bool> unclosed() const;

        /** \brief close a mesh if it wasn't closed (ensure c[end]==c[0])
            \return <the closed mesh,was it opened ?>
        */
        std::pair <MC_curve,bool> closed() const;


        /** \brief resample the curve linearly and uniformly
            \param d_L the interval length
            \return the new curve and the correspondance (int_pair_unique gives the 2 index of the original curve and the double vector corresponding to the t value)
        */
        std::pair<MC_curve,std::pair<std::vector<std::pair<MC_int_pair_unique,MC_double_vector> >,MC_double_vector> > sample_linear(const double& d_L) const;

        /** \brief get all the segments of the curve */
        std::vector <MC_segment> segment() const;
        /** \brief get the relative position of every vertex in the curve (as if the curve was parameterized uniformly between [0,1]) */
        MC_double_vector relativ_position() const;
        /** \brief return the value as if the curve was parameterized uniformly between [0,1]
            \return the corresponding value
            \return the index of the segment in which is sampled the value and its relative position to its segment
        */
        std::pair <MC_v3d,std::pair<MC_int_pair_unique,double> > value(const double& t) const;
        /** \brief return the value as if the curve was parameterized uniformly between [0,1]
            \return the corresponding value
            \return the index of the segment in which is sampled the value and there corresponding relative value to its segment
        */
        std::pair <MC_v3d_vector,std::vector<std::pair <MC_int_pair_unique,MC_double_vector> > > value(const MC_double_vector& t) const;

        /** \brief sample the curve given the current segment and the relative position in this segment
            \return the 3d position
            \return its relative position compared to the curve length
        */
        std::pair <MC_v3d,double> value(const int& k_segment,const double& relative_position) const;
        /** \brief sample the curve given the current segment and the relative position in this segment
            \return the 3d position
            \return its relative position compared to the curve length
        */
        std::pair <MC_v3d_vector,MC_double_vector> value(const MC_int_vector& k_segment,const MC_double_vector& _relative_position) const;

        /** \brief return value sampled in the t domain int(t)=segment, t-int(t)=linear interpolation
            \return the 3d position
            \return its relative position compared to the curve length
        */
        std::pair <MC_v3d_vector,MC_double_vector> value_t(const MC_double_vector t) const;

        //************************************************//
        //************************************************//
        //INTPUT/OUTPUT
        //************************************************//
        //************************************************//

        /** \brief read a stream in lin format
            \param stream of the input_file.lin
            \return the loaded Curve

            lin is obj like file

            \verbatim
            v x y z
            ...
            s n_i n_{i+1}
            ...
            \endverbatim

            indexing starts at 1 like obj
            (readable in Graphite)

            Assume only one consecutive line is given

            s should be of the form

            \verbatim
            s 0 1
            s 1 2
            ...
            \endverbatim

            Automatically avoid vertices duplication using set and V_3D comparator */
        static MC_curve read_lin(std::istream& stream);
        /** \brief load a .vect file
            \param filename the name of the file
            \return the loaded curve
            \sa read_vect(std::istream& stream)4

            Geomview can read it.

            load only one curve (return the first one if multiple curves).

            (do not return the color information)

       */
        static MC_curve read_vect_curve(std::istream& stream);
        /** \brief load one curve saved in .vect format from a file
            \sa MC_curve read_vect_curve(std::istream& stream)
        */
        static MC_curve read_vect_curve(const std::string& filename);

        /** \brief export a vect file (http://www.geomview.org/docs/oogltour.html)

            \param std::ostream& stream the output stream
            \param std::vector<MC_curve>& v_curve the input curves
            \param MC_v3d_vector v_color the index and the vector of colors (v_color.size=v_curve.size or v_color.size=0 (default))

            VECT
            N_polylines N_vertices N_color
            N_curve_0 N_curve_1 ... N_curve_n
            index_color_0 index_color_1 ... index_color_curve_n

            #vertices
            x00 y00 z00 ... x0n y0n z0n (N=N_curve_0)
            x10 y10 z10 ... x1n y1n z1n (N=N_curve_1)
            ...
            xp0 yp0 zp0 ... xnn ynn znn (N=N_curve_n, p=N_polylines)

            #color
            r0 b0 g0
            ...
            rm bm gm (m=N_color)

            Note 1 color is exported for each polyline
        */
        static void write_vect(std::ostream& stream,const std::vector<MC_curve>& v_curve,const MC_v3d_vector& v_color=MC_v3d_vector());
        /** \brief export a vector file into a given file
            \sa static void write_vect(std::ostream& stream);
        */
        static void write_vect(const std::string& filename,const std::vector<MC_curve>& v_curve,const MC_v3d_vector& v_color=MC_v3d_vector());
        /** \brief export a given curve into a vect file
            \sa static void write_vect(std::ostream& stream);
        */
        void write_vect(const std::string& filename) const;

        /** \brief read a full vect stream */
        static std::pair<std::vector<MC_curve>,MC_v3d_vector> read_vect(std::istream& stream);
         /** \brief read a full vect file */
        static std::pair<std::vector<MC_curve>,MC_v3d_vector> read_vect(const std::string& filename);


        //************************************************//
        //************************************************//
        //Diff curve
        //************************************************//
        //************************************************//

        /** \brief differenciate the position with forward euler (same size than curve) */
        MC_v3d_vector diff_forward() const;

        /** \brief differenciate the position with forward euler assuming a closed curve

            note: points should not be duplicated at the end to avoid 0 norm
        */
        MC_v3d_vector diff_forward_close() const;

        //************************************************//
        //************************************************//
        //Classical Shape
        //************************************************//
        //************************************************//


        /** \brief build a (x,y) circle of radius R, and N vertices */
        static MC_curve build_circle(const double& R,const int& N=20,const MC_v3d& normal=MC_v3d(0,0,1));

        /** \brief build a (x,y) ellipsoid of radius R1-R2, and N vertices */
        static MC_curve build_ellipsoid(const double& R1,const double& R2,const int& N=20,const MC_v3d& e0=MC_v3d(),const MC_v3d& e1=MC_v3d(),const double& theta_1=0,const double& theta_2=2*3.14159);

        /** \brief build a heart of (4N points, and circles of radius R); */
        static MC_curve build_heart(const double& R,const int& N);

        /** \brief build a line between position X0 and X1 of N samples */
        static MC_curve build_line(const MC_v3d& X0,const MC_v3d& X1,const int& N=10);

        /** \brief build a uniform B-Spline given the control-polygon */
        static MC_curve build_B_spline(const MC_curve& control_polygon,const int& N_subdiv=10);

        /** \brief build a square of 4(N-1) points of radius 1 and normal (0,0,1)*/
        static MC_curve build_square(const int& N);



        //************************************************//
        //************************************************//
        //Laplacian deformation
        //************************************************//
        //************************************************//

        /** \brief deform the curve using laplacian deformation (non closed curve) */
        MC_curve laplacian_deformation(const MC_int_vector& index_constraint,const MC_v3d_vector& position_constraint,const MC_double_vector& weight) const;

        //************************************************//
        //************************************************//
        //Curve skinning
        //************************************************//
        //************************************************//

        /** \brief skinning of curve */
        MC_curve skinning(const MC_curve& old_skeleton,const MC_curve& new_skeleton) const;


        //************************************************//
        //************************************************//
        //Rigid deformation
        //************************************************//
        //************************************************//

        /** \brief deform a closed curve with a rigid deformation (no length change) using an iterative method */
        MC_curve isometric_deformation(const MC_int_vector& constraint_index,const MC_v3d_vector& constraint_position,const MC_curve& original_curve,const double& blending_factor) const;

        private:

    };
}

#endif
