/*
**    Mesh Converter
**    Copyright (C) 2009  Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MC_DOUBLE_VECTOR_H_
#define MC_DOUBLE_VECTOR_H_

#include <stdlib.h>
#include <vector>
#include <map>
#include <set>
#include <iostream>
#include <algorithm>



namespace mesh_conv
{

    class MC_int_vector;
    class MC_v3d;
    class MC_v3d_vector;


    // helper functor class to compare struct
    class comparator_less_double_int
    {
         public:
        bool operator()(const std::pair<double,int> a,const std::pair <double,int> b) const
        {return (a.first<b.first);}
    };


  /// A container class for a vector of double
  /** internal storage is std::vector <double> */

  class MC_double_vector
  {

  public:

    //************************************************//
    //************************************************//
    //CONSTRUCTORS
    //************************************************//
    //************************************************//
    
    /** \brief standard constructor */
    MC_double_vector();
    /** \brief direct constructor with one value */
    MC_double_vector(const double& u0);
    /** \brief direct constructor with multiple values */
    MC_double_vector(const double& u0,const double& u1);
    /** \brief direct constructor with multiple values */
    MC_double_vector(const double& u0,const double& u1,const double& u2);
    /** \brief direct constructor with multiple values */
    MC_double_vector(const double& u0,const double& u1,const double& u2,const double& u3);
    /** \brief direct constructor with multiple values */
    MC_double_vector(const double& u0,const double& u1,const double& u2,const double& u3,const double& u4);
    /** \brief direct constructor with multiple values */
    MC_double_vector(const double& u0,const double& u1,const double& u2,const double& u3,const double& u4,const double& u5);
    /** \brief direct constructor from string 
	
        Can handle any string with given separator "5.4,4.5,6.1,7.8,4.1..." means <5.4,4.5,6.1,7.8,4.1,...>
     */
    MC_double_vector(const std::string& in,const std::string& separator=" ");

    /** \brief constructor from a set */
    MC_double_vector(const std::set <double>& vec);
    /** \brief constructor from an indexed map 
	
	suppose that the second value is the index of the values
     */
    MC_double_vector(const std::map <double,int> &vec);
    /** \brief constructor from a vector */
    MC_double_vector(const std::vector <double> &vec);
    /** \brief constructor from a MC_v3d */
    MC_double_vector(const MC_v3d& vec);
    /** \brief copy constructor */
    MC_double_vector(const MC_double_vector& vec);


    //************************************************//
    //************************************************//
    //Asserts
    //************************************************//
    //************************************************//

    /** \brief Assert the value stays within the bounds 
	\return nothing, but stop the programm if it goes outside bounds
     */
    void assert_bounds(const int& u) const;



    //************************************************//
    //************************************************//
    //change type
    //************************************************//
    //************************************************//

    /** \brief get the internal vector 
	\return the std::vector of the internal parameters
     */
    std::vector <double>& get_vector();
    /** \brief convert to vector (get internal vector)
        \return the std::vector corresponding to the MC_double_vector

	return the const reference of the internal parameter.
     */
    const std::vector <double>& to_vector() const;

    /** \brief convert to std::set
	\return the std::set corresponding to this vector
     */
    std::set <double> to_set() const;
    /** \brief convert to std::map
	\return the std::map corresponding to this vector (the int value return its index in the vector)
     */
    std::map <double,int> to_map() const;


    

    //************************************************//
    //************************************************//
    //Initialize value
    //************************************************//
    //************************************************//

    /** \brief build a vector of size new_size filled with zero */
    static MC_double_vector zeros(const int& new_size);

    /** \brief initialize a linspace vector */
    static MC_double_vector linspace(const double& begin,const double& end,const double& iteration=1);
    /** \brief initialize a vector with constant space and given number of sample */
    static MC_double_vector sample(const double& begin=0,const double& end=1,const int& N_sample=100);


    //************************************************//
    //************************************************//
    //manipulate
    //************************************************//
    //************************************************//

    /** \brief clear the vector to size 0*/
    void clear();

    /** \brief resize the vector in deleting value of adding zeros 
	\return the new value
     */
    MC_double_vector& resize(const int& new_size);

    /** \brief add a value at the end of the vector 
	\return the current vector 
     */
    MC_double_vector& add(const double& val);
    /** \brief add a value at the end of the vertex 
	\return the current vector (copy)
    */
    friend MC_double_vector operator<<(const MC_double_vector& vec,const double& value);


    /** \brief add some values at the end of the vector 
	\return the new current vector
     */
    MC_double_vector& add(const MC_double_vector& val);
    /** \brief add some values at the end of the vertex 
	\return the new vector (copy)
    */
    friend MC_double_vector operator<<(const MC_double_vector& vec,const MC_double_vector& to_add);

    /** \brief set an indexed value to a certain component. Adjust the size automatically 
	\return the new vector
     */
    MC_double_vector& set(const int& k_index,const double& new_value);
    /** \brief set an indexed value to a certain component. Adjust the size automatically 
	\return the new vector

	new_value and k_index must have the same size, or new_value must be of size 1
        ex. set(MC_int_vector("1 7 5 8"),MC_double_vector("4.1 -5.5 6.7 8.1"))
        or set(MC_int_vector("1 7 5 8"),MC_double_vector("-1.5"))
     */
    MC_double_vector& set(const MC_int_vector& k_index,const MC_double_vector& new_value);

    /** \brief delete the k_th value 
	\warning, this is internaly slow

	return <the new vector, the deleted value>
     */
    std::pair <MC_double_vector,double> delete_index(const int& index_to_delete) const;
    /** \brief delete the k_th value 
	\warning, this is internaly slow

	return <the new vector,the deleted value>
     */
    std::pair <MC_double_vector,MC_double_vector> delete_index(const MC_int_vector& index_to_delete) const;
    

    /** \brief ordonate the vector in increasing order
        \return the ordonated value (MC_double_vector)
        \return the corresponding mapping (MC_int_vector)
    */
    std::pair <MC_double_vector,MC_int_vector> sort() const;

    /** \brief invert values of the vector (last<->first) */
    MC_double_vector inverted() const;


    //************************************************//
    //************************************************//
    //Informations
    //************************************************//
    //************************************************//

    /** \brief return the size of the MC_double_vector */
    int size() const;

    /** \brief get min value and its position */
    std::pair <double,int> min() const;
    /** \brief get max value and its position */
    std::pair <double,int> max() const;

    //************************************************//
    //************************************************//
    //Math operator
    //************************************************//
    //************************************************//

    /** \brief change the opposite sign of the current vector (unary negation)
	\return a new vector with opposite sign
     */
    friend MC_double_vector operator-(const MC_double_vector& v);

    /** \brief add a double value to the vector
	\return the new vector 
     */
    friend MC_double_vector operator+(const MC_double_vector& vec,const double& to_add);
    /** \brief substract a double value to the vector
	\return the new vector 
     */
    friend MC_double_vector operator-(const MC_double_vector& vec,const double& to_sub);
    /** \brief multiply a double value to the vector
	\return the new vector 
     */
    friend MC_double_vector operator*(const MC_double_vector& vec,const double& to_mult);
    /** \brief multiply a double value to the vector
        \return the new vector
     */
    friend MC_double_vector operator*(const double& to_mult,const MC_double_vector& vec);

    /** \brief divide a double value to the vector
	\return the new vector 
     */
    friend MC_double_vector operator/(const MC_double_vector& vec,const double& to_subdiv);


    /** \brief multiply a double_vector to a MC_v3d
        \return the MC_v3d_vector corresponding

        pointwise multiplication
     */
    friend MC_v3d_vector operator*(const MC_double_vector& vec,const MC_v3d& to_mult);
    /** \brief multiply a double_vector to a MC_v3d
        \return the MC_v3d_vector corresponding

        pointwise multiplication
     */
    friend MC_v3d_vector operator*(const MC_v3d& to_mult,const MC_double_vector& vec);
    /** \brief divide a double value to a vector
        \return the MC_v3d_vector corresponding
     */
    friend MC_v3d_vector operator/(const MC_v3d& vec,const MC_double_vector& to_subdiv);



    /** \brief add a set of value to the vector 
	\return the new vector 

	to_add must be of size 1 or vec.size()
     */
    friend MC_double_vector operator+(const MC_double_vector& vec,const MC_double_vector& to_add);
    /** \brief substract set of value to the vector
	\return the new vector 

	to_sub must be of size 1 or vec.size()
     */
    friend MC_double_vector operator-(const MC_double_vector& vec,const MC_double_vector& to_sub);
    /** \brief multiply a set of value to the vector (pointwise mult)
	\return the new vector 

	to_mult must be of size 1 or vec.size()
     */
    friend MC_double_vector operator*(const MC_double_vector& vec,const MC_double_vector& to_mult);
    /** \brief divide a set of double value to the vector
	\return the new vector 

	to_subdiv must be of size 1 or vec.size()
     */
    friend MC_double_vector operator/(const MC_double_vector& vec,const MC_double_vector& to_subdiv);



    /** \brief internal add a value */
    MC_double_vector& operator+=(const double& to_add);
    /** \brief internal substraction */
    MC_double_vector& operator-=(const double& to_sub);
    /** \brief internal multiplication */
    MC_double_vector& operator*=(const double& to_mult);
    /** \brief internal divide */
    MC_double_vector& operator/=(const double& to_subdiv);

    /** \brief internal add values 

	to_add must be of size 1 or this->size()
     */
    MC_double_vector& operator+=(const MC_double_vector& to_add);
    /** \brief internal substraction 

	to_add must be of size 1 or this->size()
    */
    MC_double_vector& operator-=(const MC_double_vector& to_sub);
    /** \brief internal multiplication 

	to_add must be of size 1 or this->size()
     */
    MC_double_vector& operator*=(const MC_double_vector& to_mult);
    /** \brief internal divide 

	to_add must be of size 1 or this->size()
     */
    MC_double_vector& operator/=(const MC_double_vector& to_subdiv);


    /** \brief pointwise boolean comparison
        \return a MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<(const MC_double_vector& vec,const double& val);
    /** \brief pointwise boolean comparison
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<=(const MC_double_vector& vec,const double& val);
    /** \brief pointwise boolean comparison
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>(const MC_double_vector& vec,const double& val);
    /** \brief pointwise boolean comparison
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>=(const MC_double_vector& vec,const double& val);
    /** \brief pointwise boolean comparison (with an epsilon equality)
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator==(const MC_double_vector& vec,const double& val);



    /** \brief pointwise boolean comparison
        \return a MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<(const double& val,const MC_double_vector& vec);
    /** \brief pointwise boolean comparison
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<=(const double& val,const MC_double_vector& vec);
    /** \brief pointwise boolean comparison
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>(const double& val,const MC_double_vector& vec);
    /** \brief pointwise boolean comparison
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>=(const double& val,const MC_double_vector& vec);
        /** \brief pointwise boolean comparison
        \return a  MC_double_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator==(const double& val,const MC_double_vector& vec);


    //************************************************//
    //************************************************//
    // Math
    //************************************************//
    //************************************************//


    /** \brief sum the values of the double_vector */
    static double sum(const MC_double_vector& input);
    /** \brief return the absolute value componentwise */
    static MC_double_vector abs(const MC_double_vector& input);

    /** \brief return the norm 2 of the vector */
    double norm_2() const;

    /** \brief perform dot product between the current vector and an other one if size are consistent */
    double dot(const MC_double_vector& v) const;


    //************************************************//
    //************************************************//
    //Get value
    //************************************************//
    //************************************************//

    /** \brief get operator with bounds checks */
    const double& operator()(const int& index) const ;
    /** \brief get operator with bounds checks */
    double& operator()(const int& index) ;
    /** \brief get operator with bounds asserts */
    const double& operator[](const int& index) const ;
    /** \brief get operator with bounds asserts */
    double& operator[](const int& index) ;

    /** \brief get a set of value */
    MC_double_vector operator()(const MC_int_vector& index) const;
    /** \brief get a set of value */
    MC_double_vector operator[](const MC_int_vector& index) const;


    /** \brief return the first value */
    const double& first() const;
    /** \brief return the first value */
    double& first();
    /** \brief return the last value */
    const double& last() const;
    /** \brief return the last value */
    double& last();


    /** \brief get the pointer on the value
        \warning use with care */
    const double* pointer() const;
    /** \brief get the pointer on the value
        \warning use with care */
    double* pointer_unprotected();

    //************************************************//
    //************************************************//
    //OUTPUT
    //************************************************//
    //************************************************//
    
    /** \brief write to ostream the list (u0 u1 u2 ... )*/
    friend std::ostream& operator<<(std::ostream& output,const MC_double_vector& in);

    /** \brief load from istream a list

      stop at the first bad character encountered
    */
    friend MC_double_vector& operator>>(std::istream& input,MC_double_vector& vec);


  private:
    
    /** \brief the internal storage: vector of double */
    std::vector <double> double_list;

  };




}

#endif
