
/*
**    Mesh Converter
**    Copyright (C) 2008 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef _MC_DOUBLE_VECTOR_VECTOR_HPP_
#define _MC_DOUBLE_VECTOR_VECTOR_HPP_

#include <MC_double_vector.hpp>

#include <stdlib.h>
#include <vector>
#include <list>

namespace mesh_conv
{
    /** \brief Container class for MC_double_vector

        correspond to a std::vector<std::vector> > and is internaly stored as a std::vector<MC_double_vector>  */
    class MC_double_vector_vector
    {
    public:

        //************************************************//
        //************************************************//
        //CONSTRUCTORS
        //************************************************//
        //************************************************//

        /** \brief standard constructor */
        MC_double_vector_vector();
        /** \brief direct constructor from some MC_double_vector */
        MC_double_vector_vector(const MC_double_vector& v0);
        /** \brief direct constructor from some MC_double_vector */
        MC_double_vector_vector(const MC_double_vector& v0,const MC_double_vector& v1);
        /** \brief direct constructor from some MC_double_vector */
        MC_double_vector_vector(const MC_double_vector& v0,const MC_double_vector& v1,const MC_double_vector& v2);
        /** \brief direct constructor from some MC_double_vector */
        MC_double_vector_vector(const MC_double_vector& v0,const MC_double_vector& v1,const MC_double_vector& v2,const MC_double_vector& v3);

        /** \brief direct constructor from a std::list <MC_double_vector> */
        MC_double_vector_vector(const std::list <MC_double_vector>& vec);
        /** \brief direct constructor from a vector <MC_double_vector> */
        MC_double_vector_vector(const std::vector <MC_double_vector>& vec);

        //************************************************//
        //************************************************//
        //Asserts
        //************************************************//
        //************************************************//

        /** \brief Assert the value stays within the bounds
            \return nothing, but stop the programm if it goes outside bounds
        */
        void assert_bounds(const int& u) const;


        //************************************************//
        //************************************************//
        //Informations
        //************************************************//
        //************************************************//

        /** \brief get the size of the vector */
        int size() const;





        //************************************************//
        //************************************************//
        //Manipulation
        //************************************************//
        //************************************************//

        /** \brief clear the vector to size 0*/
        void clear();

        /** \brief resize the vector in deleting value of adding zeros
            \return the new value
        */
        MC_double_vector_vector& resize(const int& new_size);

        /** \brief add a value at the end of the vector
            \return the current vector
        */
        MC_double_vector_vector& add(const MC_double_vector& val);
        /** \brief add a value at the end of the vertex
            \return the current vector (copy)
        */
        friend MC_double_vector_vector operator<<(const MC_double_vector_vector& vec,const MC_double_vector& value);


        /** \brief add some values at the end of the vector
            \return the new current vector
        */
        MC_double_vector_vector& add(const MC_double_vector_vector& val);
        /** \brief add some values at the end of the vertex
            \return the new vector (copy)
        */
        friend MC_double_vector_vector operator<<(const MC_double_vector_vector& vec,const MC_double_vector_vector& to_add);

        /** \brief set an indexed value to a certain component. Adjust the size automatically
            \return the new vector
        */
        MC_double_vector_vector& set(const int& k_index,const MC_double_vector& new_value);
        /** \brief set an indexed value to a certain component. Adjust the size automatically
            \return the new vector
        */
        MC_double_vector_vector& set(const MC_int_vector& k_index,const MC_double_vector_vector& new_value);

        /** \brief delete the k_th value
            \warning, this is internaly slow

            return <the new vector, the deleted value>
        */
        std::pair <MC_double_vector_vector,MC_double_vector> delete_index(const int& index_to_delete) const;
        /** \brief delete the k_th value
            \warning, this is internaly slow

            return <the new vector,the deleted value>
        */
        std::pair <MC_double_vector_vector,MC_double_vector_vector> delete_index(const MC_int_vector& index_to_delete) const;

        /** \brief create an empty vector of size N*/
        static MC_double_vector_vector empty(const int& N);

        //************************************************//
        //************************************************//
        //Math operator
        //************************************************//
        //************************************************//

        /** \brief add an int to all the entries */
        friend MC_double_vector_vector operator+(const MC_double_vector_vector& vec,const double& to_add);
        /** \brief substract an int to all the entries */
        friend MC_double_vector_vector operator-(const MC_double_vector_vector& vec,const double& to_add);


        //************************************************//
        //************************************************//
        //Static initialization
        //************************************************//
        //************************************************//

        /** \brief repeat a MC_int_vector a certain amount of time

            ex. MC_int_vector_vector d=MC_int_vector_vector::repeat(MC_int_vector(4,5,6),3) = [[4,5,6];[4,5,6];[4,5,6]];
        */
        static MC_double_vector_vector repeat(const MC_double_vector& input,const int& N_repeat);

        //************************************************//
        //************************************************//
        //Get value
        //************************************************//
        //************************************************//

        /** \brief get operator with bounds checks */
        const MC_double_vector& operator()(const int& index) const ;
        /** \brief get operator with bounds checks */
        MC_double_vector& operator()(const int& index) ;
        /** \brief get operator with bounds asserts */
        const MC_double_vector& operator[](const int& index) const ;
        /** \brief get operator with bounds asserts */
        MC_double_vector& operator[](const int& index) ;



        /** \brief get a set of value */
        MC_double_vector_vector operator()(const MC_int_vector& index) const;
        /** \brief get a set of value */
        MC_double_vector_vector operator[](const MC_int_vector& index) const;




        /** \brief return the first value */
        const MC_double_vector& first() const;
        /** \brief return the first value */
        MC_double_vector& first();
        /** \brief return the last value */
        const MC_double_vector& last() const;
        /** \brief return the last value */
        MC_double_vector& last();

        /** \brief get pointer on internal data */
        const MC_double_vector* pointer() const;
        /** \brief get pointer on internal data
            \warning unprotected
        */
        MC_double_vector* pointer_unprotected();




        //************************************************//
        //************************************************//
        //Input/Output
        //************************************************//
        //************************************************//

        friend std::ostream& operator<<(std::ostream& stream,const MC_double_vector_vector& input);

    private:

        /** \brief internal storage as a std::vector <MC_double_vector> */
        std::vector <MC_double_vector> double_vec_list;
    };
}

#endif
