/*
**    Mesh Converter
**    Copyright (C) 2009  Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MC_INT_VECTOR_H_
#define MC_INT_VECTOR_H_

#include <stdlib.h>
#include <vector>
#include <map>
#include <set>
#include <list>
#include <iostream>

#include <MC_curve.hpp>
#include <MC_v3d.hpp>

namespace mesh_conv
{

    class MC_int_pair;
    class MC_int_pair_less;

    // helper class to compare struct
    class comparator_less_int_int
    {
         public:
        bool operator()(const std::pair<int,int> a,const std::pair <int,int> b) const
        {return (a.first<b.first);}
    };


  /// A container class for a vector of int
  /** internal storage is std::vector <int> */

  class MC_int_vector
  {

  public:

    //************************************************//
    //************************************************//
    //CONSTRUCTORS
    //************************************************//
    //************************************************//
    
    /** \brief standard constructor */
    MC_int_vector();
    /** \brief direct constructor with one value */
    MC_int_vector(const int& u0);
    /** \brief direct constructor with multiple values */
    MC_int_vector(const int& u0,const int& u1);
    /** \brief direct constructor with multiple values */
    MC_int_vector(const int& u0,const int& u1,const int& u2);
    /** \brief direct constructor with multiple values */
    MC_int_vector(const int& u0,const int& u1,const int& u2,const int& u3);
    /** \brief direct constructor with multiple values */
    MC_int_vector(const int& u0,const int& u1,const int& u2,const int& u3,const int& u4);
    /** \brief direct constructor with multiple values */
    MC_int_vector(const int& u0,const int& u1,const int& u2,const int& u3,const int& u4,const int& u5);
    /** \brief direct constructor from string 
	
	Can handle any string with given separator "5,4,6,7,4..." means <5,4,6,7,4,...>
     */
    MC_int_vector(const std::string& in,const std::string& separator=" ");

    /** \brief constructor from a set */
    MC_int_vector(const std::set <int>& vec);
    /** \brief constructor from a set */
    MC_int_vector(const std::set <unsigned int>& vec);

    /** \brief constructor from an indexed map 
	
	suppose that the second value is the index of the values
     */
    MC_int_vector(const std::map <int,int> &vec);
    /** \brief constructor from a list */
    MC_int_vector(const std::list <int> &vec);
    /** \brief constructor from a vector*/
    MC_int_vector(const std::vector <int> &vec);
    /** \brief copy constructor */
    MC_int_vector(const MC_int_vector& vec);

    //************************************************//
    //************************************************//
    //Asserts
    //************************************************//
    //************************************************//

    /** \brief Assert the value stays within the bounds 
	\return nothing, but stop the programm if it goes outside bounds
     */
    void assert_bounds(const int& u) const;



    //************************************************//
    //************************************************//
    //change type
    //************************************************//
    //************************************************//

    /** \brief get the internal vector 
	\return the std::vector of the internal parameters
     */
    std::vector <int>& get_vector();
    /** \brief convert to vector (get internal vector)
	\return the std::vector corresponding to the MC_int_vector

	return the const reference of the internal parameter.
     */
    const std::vector <int>& to_vector() const;

    /** \brief convert to std::set
	\return the std::set corresponding to this vector
     */
    std::set <int> to_set() const;
    /** \brief convert to std::map
	\return the std::map corresponding to this vector (the int value return its index in the vector)
     */
    std::map <int,int> to_map() const;
    /** \brief convert to std::list */
    std::list <int> to_list() const;

    MC_v3d to_v3d() const;
    

    //************************************************//
    //************************************************//
    //Initialize value
    //************************************************//
    //************************************************//

    /** \brief build a vector of size new_size filled with zero */
    static MC_int_vector zeros(const int& new_size);

    /** \brief linspace vector

        ex. linspace(4,8)=[4,5,6,7,8]
        ex. linspace(7,-2,-2)=[7,5,3,1,-1]
     */
    static MC_int_vector linspace(const int& begin,const int& end,const double& iteration=1);

    //************************************************//
    //************************************************//
    //manipulate
    //************************************************//
    //************************************************//

    /** \brief clear the vector to size 0*/
    void clear();

    /** \brief resize the vector in deleting value of adding zeros 
	\return the new value
     */
    MC_int_vector& resize(const int& new_size);

    /** \brief add a value at the end of the vector 
	\return the current vector 
     */
    MC_int_vector& add(const int& val);
    /** \brief add a value at the end of the vertex 
	\return the current vector (copy)
    */
    friend MC_int_vector operator<<(const MC_int_vector& vec,const int& value);


    /** \brief add some values at the end of the vector 
	\return the new current vector
     */
    MC_int_vector& add(const MC_int_vector& val);
    /** \brief add some values at the end of the vertex 
	\return the new vector (copy)
    */
    friend MC_int_vector operator<<(const MC_int_vector& vec,const MC_int_vector& to_add);

    /** \brief set an indexed value to a certain component. Adjust the size automatically 
	\return the new vector
     */
    MC_int_vector& set(const int& k_index,const int& new_value);
    /** \brief set an indexed value to a certain component. Adjust the size automatically 
	\return the new vector

	new_value and k_index must have the same size, or new_value must be of size 1
	ex. set(MC_int_vector("4 5 6 8"),MC_int_vector("1 7 -5 -8"))
	or set(MC_int_vector("4 5 6 8"),MC_int_vector("-1"))
     */
    MC_int_vector& set(const MC_int_vector& k_index,const MC_int_vector& new_value);

    /** \brief delete the k_th value 
	\warning, this is internaly slow

	return <the new vector, the deleted value>
     */
    std::pair <MC_int_vector,int> delete_index(const int& index_to_delete) const;    
    /** \brief delete the k_th value 
	\warning, this is internaly slow

	return <the new vector,the deleted value>
     */
    std::pair <MC_int_vector,MC_int_vector> delete_index(const MC_int_vector& index_to_delete) const;
    
    /** \brief ordonate the vector in increasing order
        \return the ordonated value (MC_int_vector)
        \return the corresponding mapping (MC_int_vector)
    */
    std::pair <MC_int_vector,MC_int_vector> sort() const;

    /** \brief apply a mapping to the values */
    MC_int_vector mapping(const std::map<int,int> map_to_apply) const;

    /** \brief invert the vector order (last<->first) */
    MC_int_vector inverted() const;

//    /** \brief inverse a mapping <index_0->index_1> from a map (the mapping must be a bijection) */
//    static std::map <int,int> reverse_index_mapping(const std::map<int,int>& input_mapping);

    //************************************************//
    //************************************************//
    //Build
    //************************************************//
    //************************************************//

    /** \brief build a connected list of indexed given a set of pairs */
    static std::list <MC_int_vector> connected_component(const std::set<MC_int_pair,MC_int_pair_less>& pair_set);


    //************************************************//
    //************************************************//
    //Informations
    //************************************************//
    //************************************************//

    /** \brief return the size of the MC_int_vector */
    int size() const;

    /** \brief try to find an existing value in the vector
        \return the index of the value if found, -1 if not
    */
    int find(const int& to_find) const;

    //************************************************//
    //************************************************//
    //Math operator
    //************************************************//
    //************************************************//

    /** \brief change the opposite sign of the current vector (unary negation)
	\return a new vector with opposite sign
     */
    MC_int_vector operator-() const;

    /** \brief add a int value to the vector 
	\return the new vector 
     */
    friend MC_int_vector operator+(const MC_int_vector& vec,const int& to_add);
    /** \brief substract a int value to the vector 
	\return the new vector 
     */
    friend MC_int_vector operator-(const MC_int_vector& vec,const int& to_sub);
    /** \brief multiply a int value to the vector 
	\return the new vector 
     */
    friend MC_int_vector operator*(const MC_int_vector& vec,const int& to_mult);
    /** \brief divide a int value to the vector 
	\return the new vector 
     */
    friend MC_int_vector operator/(const MC_int_vector& vec,const int& to_subdiv);



    /** \brief add a set of value to the vector 
	\return the new vector 

	to_add must be of size 1 or vec.size()
     */
    friend MC_int_vector operator+(const MC_int_vector& vec,const MC_int_vector& to_add);
    /** \brief substract a int value to the vector 
	\return the new vector 

	to_sub must be of size 1 or vec.size()
     */
    friend MC_int_vector operator-(const MC_int_vector& vec,const MC_int_vector& to_sub);
    /** \brief multiply a int value to the vector 
	\return the new vector 

	to_mult must be of size 1 or vec.size()
     */
    friend MC_int_vector operator*(const MC_int_vector& vec,const MC_int_vector& to_mult);
    /** \brief divide a int value to the vector 
	\return the new vector 

	to_subdiv must be of size 1 or vec.size()
     */
    friend MC_int_vector operator/(const MC_int_vector& vec,const MC_int_vector& to_subdiv);



    /** \brief internal add a value */
    MC_int_vector& operator+=(const int& to_add);
    /** \brief internal substraction */
    MC_int_vector& operator-=(const int& to_sub);
    /** \brief internal multiplication */
    MC_int_vector& operator*=(const int& to_mult);
    /** \brief internal divide */
    MC_int_vector& operator/=(const int& to_subdiv);

    /** \brief internal add values 

	to_add must be of size 1 or this->size()
     */
    MC_int_vector& operator+=(const MC_int_vector& to_add);
    /** \brief internal substraction 

	to_add must be of size 1 or this->size()
    */
    MC_int_vector& operator-=(const MC_int_vector& to_sub);
    /** \brief internal multiplication 

	to_add must be of size 1 or this->size()
     */
    MC_int_vector& operator*=(const MC_int_vector& to_mult);
    /** \brief internal divide 

	to_add must be of size 1 or this->size()
     */
    MC_int_vector& operator/=(const MC_int_vector& to_subdiv);


    /** \brief pointwise boolean comparison
        \return a MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<(const MC_int_vector& vec,const int& val);
    /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<=(const MC_int_vector& vec,const int& val);
    /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>(const MC_int_vector& vec,const int& val);
    /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>=(const MC_int_vector& vec,const int& val);
        /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator==(const MC_int_vector& vec,const int& val);



    /** \brief pointwise boolean comparison
        \return a MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<(const int& val,const MC_int_vector& vec);
    /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator<=(const int& val,const MC_int_vector& vec);
    /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>(const int& val,const MC_int_vector& vec);
    /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator>=(const int& val,const MC_int_vector& vec);
        /** \brief pointwise boolean comparison
        \return a  MC_int_vector with the same size as the entry (result is given by 0 or 1)
    */
    friend MC_int_vector operator==(const int& val,const MC_int_vector& vec);



    //************************************************//
    //************************************************//
    //Get value
    //************************************************//
    //************************************************//

    /** \brief get operator with bounds checks */
    const int& operator()(const int& index) const ;
    /** \brief get operator with bounds checks */
    int& operator()(const int& index) ;
    /** \brief get operator with bounds asserts */
    const int& operator[](const int& index) const ;
    /** \brief get operator with bounds asserts */
    int& operator[](const int& index) ;

    /** \brief get a set of value */
    MC_int_vector operator()(const MC_int_vector& index) const;
    /** \brief get a set of value */
    MC_int_vector operator[](const MC_int_vector& index) const;


    /** \brief return the first value */
    const int& first() const;
    /** \brief return the first value */
    int& first();
    /** \brief return the last value */
    const int& last() const;
    /** \brief return the last value */
    int& last();

    /** \brief internal delete the last value */
    MC_int_vector& erase_last();


    /** \brief get the pointer on the data */
    const int* pointer() const;

    //************************************************//
    //************************************************//
    //OUTPUT
    //************************************************//
    //************************************************//
    
    /** \brief write to ostream the list (u0 u1 u2 ... )*/
    friend std::ostream& operator<<(std::ostream& output,const mesh_conv::MC_int_vector& in);

    /** \brief load from istream a list

      stop at the first bad character encountered
    */
    friend mesh_conv::MC_int_vector& operator>>(std::istream& input,mesh_conv::MC_int_vector& vec);


  private:
    
    /** \brief the internal storage: vector of int */
    std::vector <int> int_list;

  };




}

#endif
