/*
**    Mesh Converter
**    Copyright (C) 2008 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MC_INT_VECTOR_VECTOR_H_
#define MC_INT_VECTOR_VECTOR_H_

#include <MC_int_vector.hpp>

#include <stdlib.h>

namespace mesh_conv
{

    class MC_int_list_vector;

    /** \brief Container class for MC_int_vector

        correspond to a std::vector<std::vector> > and is internaly stored as a std::vector<MC_int_vector>  */
    class MC_int_vector_vector
    {
    public:

    //************************************************//
    //************************************************//
    //CONSTRUCTORS
    //************************************************//
    //************************************************//

    /** \brief standard constructor */
    MC_int_vector_vector();
    /** \brief direct constructor from some MC_int_vector */
    MC_int_vector_vector(const MC_int_vector& v0);
    /** \brief direct constructor from some MC_int_vector */
    MC_int_vector_vector(const MC_int_vector& v0,const MC_int_vector& v1);
    /** \brief direct constructor from some MC_int_vector */
    MC_int_vector_vector(const MC_int_vector& v0,const MC_int_vector& v1,const MC_int_vector& v2);
    /** \brief direct constructor from some MC_int_vector */
    MC_int_vector_vector(const MC_int_vector& v0,const MC_int_vector& v1,const MC_int_vector& v2,const MC_int_vector& v3);

    /** \brief constructor from a MC_int_list_vector */
    MC_int_vector_vector(const MC_int_list_vector& v);

    /** \brief direct constructor from a std::list <MC_int_vector> */
    MC_int_vector_vector(const std::list <MC_int_vector>& vec);

    //************************************************//
    //************************************************//
    //Asserts
    //************************************************//
    //************************************************//

    /** \brief Assert the value stays within the bounds
        \return nothing, but stop the programm if it goes outside bounds
     */
    void assert_bounds(const int& u) const;


    //************************************************//
    //************************************************//
    //Informations
    //************************************************//
    //************************************************//

    /** \brief get the size of the vector */
    int size() const;





    //************************************************//
    //************************************************//
    //Manipulation
    //************************************************//
    //************************************************//

    /** \brief clear the vector to size 0*/
    void clear();

    /** \brief resize the vector in deleting value of adding zeros
        \return the new value
     */
    MC_int_vector_vector& resize(const int& new_size);
    /** \brief create an empty container of a given size */
    static MC_int_vector_vector empty(const int& N);

    /** \brief add a value at the end of the vector
        \return the current vector
     */
    MC_int_vector_vector& add(const MC_int_vector& val);
    /** \brief add a value at the end of the vertex
        \return the current vector (copy)
    */
    friend MC_int_vector_vector operator<<(const MC_int_vector_vector& vec,const MC_int_vector& value);


    /** \brief add some values at the end of the vector
        \return the new current vector
     */
    MC_int_vector_vector& add(const MC_int_vector_vector& val);
    /** \brief add some values at the end of the vertex
        \return the new vector (copy)
    */
    friend MC_int_vector_vector operator<<(const MC_int_vector_vector& vec,const MC_int_vector_vector& to_add);

    /** \brief set an indexed value to a certain component. Adjust the size automatically
        \return the new vector
     */
    MC_int_vector_vector& set(const int& k_index,const MC_int_vector& new_value);
    /** \brief set an indexed value to a certain component. Adjust the size automatically
        \return the new vector
     */
    MC_int_vector_vector& set(const MC_int_vector& k_index,const MC_int_vector_vector& new_value);

    /** \brief delete the k_th value
        \warning, this is internaly slow

        return <the new vector, the deleted value>
     */
    std::pair <MC_int_vector_vector,MC_int_vector> delete_index(const int& index_to_delete) const;
    /** \brief delete the k_th value
        \warning, this is internaly slow

        return <the new vector,the deleted value>
     */
    std::pair <MC_int_vector_vector,MC_int_vector_vector> delete_index(const MC_int_vector& index_to_delete) const;


    //************************************************//
    //************************************************//
    //Math operator
    //************************************************//
    //************************************************//

    /** \brief add an int to all the entries */
    friend MC_int_vector_vector operator+(const MC_int_vector_vector& vec,const int& to_add);
    /** \brief substract an int to all the entries */
    friend MC_int_vector_vector operator-(const MC_int_vector_vector& vec,const int& to_add);


    //************************************************//
    //************************************************//
    //Static initialization
    //************************************************//
    //************************************************//

    /** \brief repeat a MC_int_vector a certain amount of time

      ex. MC_int_vector_vector d=MC_int_vector_vector::repeat(MC_int_vector(4,5,6),3) = [[4,5,6];[4,5,6];[4,5,6]];
    */
    static MC_int_vector_vector repeat(const MC_int_vector& input,const int& N_repeat);

    //************************************************//
    //************************************************//
    //Get value
    //************************************************//
    //************************************************//

    /** \brief get operator with bounds checks */
    const MC_int_vector& operator()(const int& index) const ;
    /** \brief get operator with bounds checks */
    MC_int_vector& operator()(const int& index) ;
    /** \brief get operator with bounds asserts */
    const MC_int_vector& operator[](const int& index) const ;
    /** \brief get operator with bounds asserts */
    MC_int_vector& operator[](const int& index) ;

    /** \brief get operator from a vector of index */
    MC_int_vector_vector operator()(const MC_int_vector& index) const;


    /** \brief return the first value */
    const MC_int_vector& first() const;
    /** \brief return the first value */
    MC_int_vector& first();
    /** \brief return the last value */
    const MC_int_vector& last() const;
    /** \brief return the last value */
    MC_int_vector& last();

    /** \brief get pointer on internal data */
    const MC_int_vector* pointer() const;
    /** \brief get pointer on internal data
        \warning unprotected
    */
    MC_int_vector* pointer_unprotected();


    //************************************************//
    //************************************************//
    //Input/Output
    //************************************************//
    //************************************************//

    friend std::ostream& operator<<(std::ostream& stream,const MC_int_vector_vector& input);

    private:

    /** \brief internal storage as a std::vector <MC_int_vector> */
    std::vector <MC_int_vector> int_vec_list;
};

}


#endif
