/*
**    Mesh Converter
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef   	MC_MATRIX_N_H_
# define   	MC_MATRIX_N_H_


#include <MC_double_vector.hpp>

#include <iostream>



namespace mesh_conv
{

    class MC_int_vector;
    class MC_v3d;
    class MC_v4d;
    class MC_v3d_vector;


    /** \brief exception class for matrix */





    /// A Full Matrix class
    /** internal access is given by M(i,j)=M[i+j*size_i]
        Many internal operations are actually based on the newmat lib http://www.robertnz.net/nm_intro.htm
     **/
    class MC_matrix
    {
    public:


        //************************************************//
        //************************************************//
        //CONSTRUCTORS
        //************************************************//
        //************************************************//

        /** \brief empty constructor */
        MC_matrix();
        /** \brief zeros matrix with given size */
        MC_matrix(const int& size_col,const int& size_row);
        /** \brief zeros square matrix with given size */
        MC_matrix(const int& size_col);
        /** \brief copy constructor */
        MC_matrix(const MC_matrix& size_row);
        /** \brief direct constructor from a MC_v3d */
        MC_matrix(const MC_v3d& v);
        /** \brief direct constructor from a MC_v3d */
        MC_matrix(const MC_v3d& v1,const MC_v3d& v2);
        /** \brief direct constructor from a MC_v3d */
        MC_matrix(const MC_v3d& v1,const MC_v3d& v2,const MC_v3d& v3);
        /** \brief direct constructor from a set of MC_v3d */
        MC_matrix(const MC_v3d_vector& v);
        /** \brief direct constructor from a column vector */
        MC_matrix(const MC_double_vector& column);


        /** \brief destructor */
        ~MC_matrix();



        //************************************************//
        //************************************************//
        //Size
        //************************************************//
        //************************************************//

        /** \brief check if size are consistent */
        /** \return nothing but exit if something is wrong */
        void check_integrity() const;

        /** resize to 0 */
        MC_matrix& clear();

        /** get size 1 */
        int size_1() const;
        /** get size 2 */
        int size_2() const;
        /** get both sizes */
        MC_int_vector size() const;
        /** get size for a vector */
        int size_vec() const;


        /** \brief internal resize the matrix */
        MC_matrix& resize(const MC_int_vector& new_size) ;
        /** \brief internal resize the matrix (if size_2==-1 by default, resize a square matrix) */
        MC_matrix& resize(const int& size_1,const int& size_2=-1) ;
        /** \brief internal resize only in x */
        MC_matrix& resize_1(const int& size_1) ;
        /** \brief internal resize only in y */
        MC_matrix& resize_2(const int& size_2) ;


        /** \brief reshape the matrix with a constant total size */
        MC_matrix reshaped_1(const int& size_1) const;
        /** \brief reshape the matrix with a constant total size */
        MC_matrix reshaped_2(const int& size_2) const;
        /** \brief reshape the matrix with a constant total size */
        MC_matrix reshaped(const MC_int_vector& new_size) const;
        /** \brief reshape the matrix with a constant total size */
        MC_matrix reshaped(const int& new_size_1,const int& new_size_2) const;

        /** \brief repeat matrix by block

            repeat k_reapeat times in x and y */
        MC_matrix repmat(const int& k_repeat) const ;
        /** \brief repeat matrix by block */
        MC_matrix repmat(const int& k_repeat_1,const int& k_repeat_2) const ;
        /** \brief repeat matrix by block */
        MC_matrix repmat(const MC_int_vector& k_repeat) const ;
        /** \brief repeat matrix by block in x direction */
        MC_matrix repmat_1(const int& k_repeat_1) const ;
        /** \brief repeat matrix by block in y direction */
        MC_matrix repmat_2(const int& k_repeat_2) const ;

    

        //************************************************//
        //************************************************//
        //Access
        //************************************************//
        //************************************************//


        /** \brief set a block of matrix

            block size must correspond to size of index_1 and 2 */
        MC_matrix& set_block(const MC_int_vector& index_1,const MC_int_vector& index_2,const MC_matrix& block);
        /** \brief set a block of matrix

            block size must correspond to size of index_1 and 2 */
        MC_matrix& set_block(const int& start_x,const int& end_x,const int& start_y,const int& end_y,const MC_matrix& block);

        /** \brief return a double if size are ok (size_1()==1 && size_2()==1) */
        const double& to_double() const;
        /** \brief return a vector of the matrix (similar to M(:)) */
        const MC_double_vector& to_vec () const;

        /** \brief add a block of MC_v3d only if size if (n,3) */
        MC_matrix& add_row(const MC_v3d& x0) ;
        /** \brief add a block of MC_v3d only if size if (n,3) */
        MC_matrix& add_row(const MC_v3d& x0,const MC_v3d& x1) ;
        /** add a block of MC_v3d only if size if (n,3) */
        MC_matrix& add_row(const MC_v3d& x0,const MC_v3d& x1,const MC_v3d& x2);
        /** \brief add a column block only if size are compatibles */
        MC_matrix& add_row(const MC_double_vector& col);
        /** \brief set a full row given a double_vector
            resize the matrix if needed
        */
        MC_matrix& set_row(const int& id_row,const MC_double_vector& row);
        /** \brief get a given row */
        MC_double_vector get_row(const int& k_index) const;



        /** \brief add a block of V_3D only if size if (3,n) */
        MC_matrix& add_col(const MC_v3d& x0);
        /** \brief add a block of V_3D only if size if (3,n) */
        MC_matrix& add_col(const MC_v3d& x0,const MC_v3d& x1);
        /** \brief add a block of V_3D only if size if (3,n) */
        MC_matrix& add_col(const MC_v3d& x0,const MC_v3d& x1,const MC_v3d& x2);
        /** \brief add a column block only if size are compatibles */
        MC_matrix& add_col(const MC_double_vector& col);

        /** \brief set a full column given a double_vector

            resize the matrix if needed
        */
        MC_matrix& set_col(const int& id_col,const MC_double_vector& col);
        /** \brief get a given row */
        MC_double_vector get_col(const int& k_index) const;

        /** \brief transform a 3x3 affine matrix into a 4x4 projective one (M[4][4]=1) */
        MC_matrix to_matrix4() const;
        /** \brief transform a 4x4 projective matrix transform into a 3x3 by resizing
            \return <MC_matrix: 3x3 matrix ,MC_v3d the translation> */
        std::pair <MC_matrix,MC_v3d> to_matrix3() const;


        //************************************************//
        //************************************************//
        //INIT
        //************************************************//
        //************************************************//

        /** \brief build a zeros matrix */
        static MC_matrix zeros(const int& size_1,const int& size_2);
        /** \brief build a square zeros matrix */
        static MC_matrix zeros(const int& size);

        /** \brief build a square identity matrix */
        static MC_matrix identity(const int& size);

        /** \brief build a rotation matrix transforming an axis a1 into an axis a2

            Get rotation matrix R such that R a1 = a2
            Automaticaly works on normalized axis
        */
        static MC_matrix rotation_axis_to_axis(const MC_v3d& a1,const MC_v3d& a2);

        /** \brief get the best rotation between a set of points

            Use registration method to find rotation:
            Let's X, X0 be the 3D position in a matrix form
            Find R such that R X = X0 in the best way for froebenius norm.
            R = S V^t, with S.D.V^t = X X0^{t}. **/
        static MC_matrix rotation_registration(const MC_matrix& X,const MC_matrix& X0);
  
        /** \brief init to a rotation matrix

            axis of the rotation (normalized inside)
            angle in radians */
        static MC_matrix rotation(const MC_v3d& axis,const double& angle);
        /** \brief get the rotation matrix transforming a0 to a1 and the associated gradient
            \return the rotation matrix R
            \return the gradient (dR/da0[0],dR/da0[1],dR/da0[2],dR/da1[0],dR/da1[1],dR/da1[2])
            \warning (a0,a1) have to be normalized before being used
        */
        static std::pair <MC_matrix,std::vector<MC_matrix> > rotation_axis_to_axis_with_gradient(const MC_v3d& a0,const MC_v3d& a1);

        /** \brief create translation matrix */
        static MC_matrix translation(const MC_v3d& tr);

        /** \brief create scaling matrix */
        static MC_matrix scaling(const double& s);
        /** \brief create scaling matrix */
        static MC_matrix scaling(const double& sx,const double& sy,const double& sz);

        //************************************************//
        //************************************************//
        //Math operator
        //************************************************//
        //************************************************//
  
        /** \brief transpose the matrix */
        MC_matrix transposed() const;

        /** \brief Invert the matrix if possible
            Use newmat class to invert */
        MC_matrix inverted() const;



        /** \brief get the norm_2 of the matrix
            return the sqrt(sum x_i^2) for every elements */
        double norm_2() const;



  

        /** \brief return the log matrix if exists
            See kenney and Laub [1989] and Alexa's paper on linear interpolation */
        MC_matrix log_m(int *error_bit=0) const;
//        /** \brief return the log matrix if exists without the error_bit
//            \sa MC_matrix log_m(int *error_bit) const; */
//        MC_matrix log_m() const;

        /** power of the matrix (with an integer) */
        MC_matrix pow_m(const int& k_pow) const;

        /** \brief get the exponential matrix

             see Golub abd Van Loan [1989] and Alexa's paper on linear interpolation */
        MC_matrix exp_m() const;


        /** \brief invert matrix in a least square sense
            \return (M^t M)^{-1) M^t

            use newmat invert
        */
        MC_matrix lsqr_invert() const;

        /** \brief get polar decomposition of the matrix [R,U] with M=R.U
            \return return M=R.U using [Higham 86] iteration method
            (faster than SVD decomposition : M=SDV^t => R=SV^t)
        */
        std::pair<MC_matrix,MC_matrix> polar_decomposition() const;

        /** \brief vector polar decomposition on multiple matrices
            \sa std::pair<MC_matrix,MC_matrix> polar_decomposition() const;
            \return the set of <Rotation,U_value>
        */
        static std::pair<std::vector <MC_matrix>,std::vector <MC_matrix> > polar_decomposition(const std::vector <MC_matrix>& v_m);


        /** \brief get trace */
        double trace() const;


        /** \brief compute kronecker product between two vectors */
        static MC_matrix kronecker(const MC_double_vector& v1,const MC_double_vector& v2);



        /** \brief componentwise product
            \warning matrices must have the same size */
        MC_matrix componentwise(const MC_matrix& M1) const;




        /** \brief internal add a scalar value */
        MC_matrix& operator+=(const double& a);
        /** \brief internal add an other matrix */
        MC_matrix& operator+=(const MC_matrix& M);
        /** \brief internal multiplication by a scalar */
        MC_matrix& operator*=(const double& a);
        /** \brief internal division by a scalar */
        MC_matrix& operator/=(const double& a);



        /** \brief add a double to every value */
        friend MC_matrix operator+(const MC_matrix& M1,const double& alpha);
        /** \brief add a double to every value */
        friend MC_matrix operator+(const double& alpha,const MC_matrix& M1);
        /** \brief add a matrix to an other (size must be compatible) */
        friend MC_matrix operator+(const MC_matrix& M1,const MC_matrix& M2);


        /** \brief substract a double to every value */
        friend MC_matrix operator-(const MC_matrix& M1,const double& alpha);
        /** \brief substract a double to every value */
        friend MC_matrix operator-(const double& alpha,const MC_matrix& M1);
        /** \brief substract a matrix to an other (size must be compatible) */
        friend MC_matrix operator-(const MC_matrix& M1,const MC_matrix& M2);

        /** \brief multiply a double to every value */
        friend MC_matrix operator*(const MC_matrix& M1,const double& alpha);
        /** \brief multiply a double to every value */
        friend MC_matrix operator*(const double& alpha,const MC_matrix& M1);
        /** \brief multiply two matrices (size must be compatible) */
        friend MC_matrix operator*(const MC_matrix& M1,const MC_matrix& M2);
        /** \brief apply a matrix to a MC_v3d vector (size_2() must be 3) */
        friend MC_matrix operator*(const MC_matrix& M1,const MC_v3d& V);
        /** \brief apply a matrix to a MC_double_vector if size_2()==double_vec.size() */
        friend MC_double_vector operator*(const MC_matrix& M1,const MC_double_vector& v);

        /** \brief internal multiplication beween two matrix 
            \warning not optimized implementation
        */
        MC_matrix& operator*=(const MC_matrix& M1);



        /** \brief divide a double to every value */
        friend MC_matrix operator/(const MC_matrix& M1,const double& alpha);
        /** \brief divide a double to every value */
        friend MC_matrix operator/(const double& alpha,const MC_matrix& M1);

        /** \brief internal product to a MC_v3d
            \sa MC_v3d::operator*=(const MC_matrix& M)
            \return nothing and the pointer to_multiply is set to the value M*to_multiply
            to goal of this function is do perform relativly fastly the product x=M*x <=> x*=M;
        */
        void internal_product(MC_v3d* to_multiply) const;
        /** \brief internal product to a MC_v4d
            \sa MC_v4d::operator*=(const MC_matrix& M)
            \return nothing and the pointer to_multiply is set to the value M*to_multiply
            to goal of this function is do perform relativly fastly the product x=M*x <=> x*=M;
        */
        void internal_product(MC_v4d* to_multiply) const;

        /** \brief internal add to the translation part of a matrix 4x4 */
        MC_matrix& add_translation(const MC_v3d& tr);

        //************************************************//
        //************************************************//
        //Get operator
        //************************************************//
        //************************************************//
  
        /** \brief direct access */
        double& operator()(const int& k1,const int& k2);
        /** \brief direct access */
        const double& operator()(const int& k1,const int& k2) const;
        /** \brief access from real index in vector

            Calculated as k1+N_1*k2 */
        double& operator()(const int& k_index);
        /** \brief access from real index in vector

            Calculated as k1+N_1*k2 */
        const double& operator()(const int& k_index) const;
        /** \brief access from vector of real index in vector
            \return a row matrix
            Calculated as k1+N_1*k2 */
        MC_matrix operator()(const MC_int_vector& k_index);

        /** \brief submatrix from vector of coordinates

            directly take the set of index_1 x index_2 as a tensor product */
        MC_matrix operator()(const MC_int_vector& index_1,const MC_int_vector& index_2) const;
        /** \brief submatrix from vector of coordinates

            take the set as a tensor product */
        MC_matrix operator()(const std::string& string_index_1,const std::string& string_index_2);
        /** \brief submatrix from a vector of coordinates given by a string
            \return a row matrix */
        MC_matrix operator()(const std::string& string_index);

        /** \brief get the translation part of a matrix 4x4
            \return get_block("0:2","3") if the matrix is 4x4 */
        MC_v3d translation_part() const;

        /** \brief set the translation part of a matrix 4x4 */
        MC_matrix& set_translation(const MC_v3d& tr);

        /** \brief set the rotation (block 3x3) part of a matrix 4x4 or 3x3 */
        MC_matrix& set_rotation(const MC_matrix& m);

        /** \brief get the pointer on the value
            \warning use with care */
        const double* pointer() const;

        //************************************************//
        //************************************************//
        //Input / Output
        //************************************************//
        //************************************************//


        /** \brief output stream */
        friend std::ostream& operator << (std::ostream& stream,const MC_matrix& _M);

        /** \brief export matrix as string */
        std::string to_string() const;

        /** \brief read transformation as an string string */
        static MC_matrix transformation(const std::string& input);


        //************************************************//
        //************************************************//
        //should go in tensor
        //************************************************//
        //************************************************//

        //2d Tensor
        static std::pair<MC_v3d,MC_v3d> tensor_to_axes(const MC_matrix& T);
        //3rd value must be 0
        static MC_matrix axes_to_tensor(const MC_v3d& e0,const MC_v3d& e1);
  

 private:

  /** \brief internal string analyser for type "start:increment:end"
      fill start,increment and end for the first row (ex. "0:2:end" or "end-15:7:end-8")
      Recognisze key word ":"; "end"; "-" */
  void analyse_string(const std::string& to_analyse,int dim,int *start,double *increment,int *end) const;
  /** \brief internal string analyser for type "start:increment:end" */
  void analyse_string(const char* to_analyse,int dim,int *start,double *increment,int *end) const;


  protected:

  /** \brief the internal storage format is a vector of double */
  MC_double_vector M;

  /** \brief the stored size (size_y x size_x)=(nbr_line x nbr_row)
      At every moment size(M)=N[0]*N[1] should be respected **/
  int current_size[2];

};

}


#endif 	    /* !MC_matrix_H_ */
