
/*
**    Mesh Converter
**    Copyright (C) 2008 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MC_MESH_FAST_DRAW_HPP_
#define MC_MESH_FAST_DRAW_HPP_

#include <MC_int_vector.hpp>
#include <MC_double_vector.hpp>

#include <vector>
#include <set>


namespace mesh_conv
{

    class MC_connectivity_index;
    class MC_v3d_vector;
    class MC_mesh_index_vector;
    class MC_mesh_index_vector_draw;

    /** \brief Mesh class for fast draw in OpenGL*/
    class MC_mesh_fast_draw
    {

    public:

        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \brief empty constructor */
        MC_mesh_fast_draw();
        /** \brief direct constructor from a point set and a connectivity */
        MC_mesh_fast_draw(const MC_connectivity_index& connectivity,const MC_v3d_vector& point_set);
        /** \brief direct constructor from a MC_mesh_index_vector */
        MC_mesh_fast_draw(const MC_mesh_index_vector& mesh);
        /** \brief direct constructor from a MC_mesh_index_vector_draw */
        MC_mesh_fast_draw(const MC_mesh_index_vector_draw& mesh);


        /** \brief classical destructor */
        ~MC_mesh_fast_draw();


        //*********************************************//
        //*********************************************//
        // get internal
        //*********************************************//
        //*********************************************//

        /** \brief get the point set */
        const MC_double_vector& point_set() const;
        /** \brief get the point set */
        MC_double_vector& point_set();
        /** \brief get the color vector */
        const MC_double_vector& color() const;
        /** \brief get the texture vector */
        const MC_double_vector& texture() const;
        /** \brief get the normal */
        const MC_double_vector& normal() const;
        /** \brief get the normal */
        MC_double_vector& normal();
        /** \brief get the connectivity */
        const std::map <int,MC_int_vector>& connectivity() const;


        //*********************************************//
        //*********************************************//
        // set from v3d class to continuous internal storage
        //*********************************************//
        //*********************************************//

        /** \brief set directly the point set from a v3d_vector */
        MC_mesh_fast_draw& set_point_set(const MC_v3d_vector& point_set);
        /** \brief set directly the connectivity set from a MC_connectivity */
        MC_mesh_fast_draw& set_connectivity(const MC_connectivity_index& connectivity);
        /** \brief set directly the color set from a v3d_vector of size vertex_number (per vertex) */
        MC_mesh_fast_draw& set_color(const MC_v3d_vector& color);
        /** \brief set directly the normal set from a v3d_vector of size vertex_number (per vertex) */
        MC_mesh_fast_draw& set_normal(const MC_v3d_vector& normal);
        /** \brief set directly the texture from a v3d_vector of size vertex_number (per vertex) */
        MC_mesh_fast_draw& set_texture(const MC_v3d_vector& texture);


    private:

        /** \brief position vector*/
        MC_double_vector point_set_mesh;
        /** \brief color vector */
        MC_double_vector color_mesh;
        /** \brief  texture vector */
        MC_double_vector texture_mesh;
        /** \brief normal vector */
        MC_double_vector normal_mesh;

        /** \brief connectivity for different kind of polygons (size,connectivity)
            \warning the connectivity has to be shared between position+color+texture
        */
        std::map <int,MC_int_vector> connectivity_mesh;

    };
}

#endif
