/*
**    Mesh Converter
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MC_MESH_INDEX_VECTOR_H_
#define MC_MESH_INDEX_VECTOR_H_


#include <MC_connectivity_index.hpp>
#include <MC_v3d_vector.hpp>

#include <stdlib.h>
#include <set>


namespace mesh_conv
{
    class MC_polygon;
    class MC_segment;
    class MC_matrix;
    class MC_double_vector_vector;
    class MC_int_pair;
    class MC_int_pair_less;

    namespace mesh
    {
        namespace cgal
        {
            class MC_mesh_cgal;
        }
    }

    /// mesh class containing connectivity and point_set based on vector
    class MC_mesh_index_vector
    {
    public:


        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \brief empty constructor */
        MC_mesh_index_vector();
        /** \brief constructor from a point set and a connectivity */
        MC_mesh_index_vector(const MC_v3d_vector& _point_set,const MC_connectivity_index& _connectivity);
        /** \brief constructor from a pair of point set and a connectivity */
        MC_mesh_index_vector(const std::pair<MC_v3d_vector,MC_connectivity_index>& _mesh);
        /** \brief constructor from a vector of polygon */
        MC_mesh_index_vector(const std::vector<MC_polygon> poly_soup);
        /** \brief constructor from a cgal mesh */
        MC_mesh_index_vector(const mesh::cgal::MC_mesh_cgal& mesh);

        //*********************************************//
        //*********************************************//
        // Informations
        //*********************************************//
        //*********************************************//

        /** \brief Number of polygons */
        int polygon_number() const;
        /** \brief Number of vertices */
        int vertex_number() const;

        /** \brief return the size of a given polygon */
        int polygon_size(const int& k_polygon) const;

        //*********************************************//
        //*********************************************//
        // Add index and vertices
        //*********************************************//
        //*********************************************//

        /** \brief add a vertex into the mesh
            \return the current mesh
        */
        MC_mesh_index_vector& add_vertex(const MC_v3d& v);
        /** \brief add an index of polygon in the connectivity
            \return the current mesh
        */
        MC_mesh_index_vector& add_connectivity_index(const MC_int_vector& c);


        /** \brief concatenation of the point set and the connectivity
            \return a new mesh in directly adding the vertices

            the connectivity is automatically incremented to avoid conflict
        */
        friend MC_mesh_index_vector operator<<(const MC_mesh_index_vector& vec0,const MC_mesh_index_vector& vec1);

        /** \brief internal concatenation of the point set and connectivity
            \return a new mesh in directly adding the vertices

            the connectivity is automatically incremented to avoid conflict
        */
        MC_mesh_index_vector& concatenation(const MC_mesh_index_vector& vec1);


        //*********************************************//
        //*********************************************//
        // Get internal structures
        //*********************************************//
        //*********************************************//

        /** \brief return the internal connectivity */
        MC_connectivity_index& connectivity();
        /** \brief return the internal connectivity */
        const MC_connectivity_index& connectivity() const;
        /** \brief return the internal point_set */
        MC_v3d_vector& point_set();
        /** \brief return the internal point_set */
        const MC_v3d_vector& point_set() const;



        //*********************************************//
        //*********************************************//
        // get
        //*********************************************//
        //*********************************************//

        /** \brief get the designated polygon */
        MC_polygon get_polygon(const int& k_index) const;

        /** \brief get the polygon soup */
        std::vector<MC_polygon> get_polygon() const;

        /** \brief return a mesh made of polygon soup */
        MC_mesh_index_vector get_polygon_mesh() const;

        //*********************************************//
        //*********************************************//
        // Subdivision
        //*********************************************//
        //*********************************************//

        /** \brief subdivision_mid_edge */
        MC_mesh_index_vector subdivide_mid_edge() const;

        /** \brief subdivision_barycenter_mid_edge */
        MC_mesh_index_vector subdivide_barycenter_mid_edge() const;

        /** \brief subdivision_mid_edge without modifying the boundaries */
        MC_mesh_index_vector subdivide_mid_edge_unchanged_boundary() const;
        /** \brief subdivision_barycenter_mid_edge without modifying the boundaries */
        MC_mesh_index_vector subdivide_barycenter_mid_edge_unchanged_boundary() const;
        /** \brief subdivision_mixed_mid_edge without modifying the boundaries */
        MC_mesh_index_vector subdivide_mixed_mid_edge_unchanged_boundary() const;

        /** \brief subdivision_mid_edge without modifying the boundaries */
        MC_mesh_index_vector subdivide_mid_edge_unchanged_boundary(const std::set<MC_int_pair,MC_int_pair_less>& allowed_edge) const;
        /** \brief subdivision_barycenter_mid_edge without modifying the boundaries */
        MC_mesh_index_vector subdivide_barycenter_mid_edge_unchanged_boundary(const std::set<MC_int_pair,MC_int_pair_less>& allowed_edge) const;//=std::set<MC_int_pair,MC_int_pair_less>) const;
        /** \brief subdivision_mixed_mid_edge without modifying the boundaries */
        MC_mesh_index_vector subdivide_mixed_mid_edge_unchanged_boundary(const std::set<MC_int_pair,MC_int_pair_less>& allowed_edge) const;//=std::set<MC_int_pair,MC_int_pair_less>) const;

        /** \brief subdivision_mixed_mid_edge of a selected number of polygons without change of boundaries
            \note slow, but could be speed-up using MC_connectivity_list
        */
        MC_mesh_index_vector subdivide_mixed_mid_edge(const std::set<int>& polygon_to_subdivide,const std::set <MC_int_pair,MC_int_pair_less>& allowed_edge) const;
        /** \brief subdivision_mixed_mid_edge of a selected number of polygons without change of boundaries
            \note slow, but could be speed-up using MC_connectivity_list
        */
        MC_mesh_index_vector subdivide_mixed_mid_edge(const std::set<int>& polygon_to_subdivide) const;


        //*********************************************//
        //*********************************************//
        // Special Shapes
        //*********************************************//
        //*********************************************//

        /** \brief build a unit cube */
        static MC_mesh_index_vector build_cube();

        /** \brief return a vector of meshes used for the local basis */
        static std::vector <MC_mesh_index_vector> build_local_basis();


        /** \brief build grid from a vector of segment */
        static MC_mesh_index_vector build_segment(const std::vector<MC_segment>& v_seg,const double& radius,const double& N_circular);

        /** \brief build a unitary icosahedron */
        static MC_mesh_index_vector build_icosahedron();
        /** \brief build unitary sphere by subdividing a icosahedron and projecting the new points */
        static MC_mesh_index_vector build_sphere(const int& N_subdiv=2);
        /** build unitary sphere by subdividing a cube and projecting the new points */
        static MC_mesh_index_vector build_quad_sphere(const int& N_subdiv=2);
        /** \brief build a parametric sphere

             Do quads, but these are not real quads */
        static MC_mesh_index_vector build_parametric_sphere(const int& N_1=10,const int& N_2=10);
        /** \brief build a torus of radius (R0,R1)
            \param R0 is the medium radius, and R1 is the small one

            Do non planar quad faces */
        static MC_mesh_index_vector build_torus(const double& R0,const double& R1,const int& N_1,const int& N_2);
        /** \brief build cylinder (add center points at the extremities)

            Radius is set to 1. Do quads. */
        static MC_mesh_index_vector build_closed_cylinder(const int& N1,const int& N2,const bool& is_closed=true);
        /** \brief build a unit cone */
        static MC_mesh_index_vector build_cone(const int& N_radius,const int& N_face);
        /** \brief build a disc
          \param center: the center of the disc (default (0,0,0))
          \param normal: the normal to the disc (default 1)
          \param radius: the radius of the disc (default 1)
          \param N_radius: the number of samples around the disc (default 10)
          \param N_interior: the number of vertex in the interior (default 2=no interior vertex)
        */
        static MC_mesh_index_vector build_disc(const MC_v3d& center=MC_v3d(0,0,0),const MC_v3d& normal=MC_v3d(0,0,1),const double& radius=1,const int& N_radius=10,const int& N_interior=2);
        /** \brief build an arrow
            \param gives the segment for the extreme points, the radius of the cylinder and the radius of the cone and its length */
        static MC_mesh_index_vector build_arrow(const MC_segment& dir,const double& radius_cylinder=0.1,const int& N_cylinder=30,const int& N_shape_cylinder=5,const double& radius_cone=0.25,const double& length_cone=0.35,const int& N_cone=30,const int& N_shape_cone=5);


        /** \brief build a strip planar plane following the curve */
        static std::pair <MC_mesh_index_vector,std::pair<MC_v3d_vector,std::vector<MC_int_pair> > > build_strip_planar(const MC_curve& c,const MC_v3d& normal,const int& number_of_lines=5,const double& binormal_length=0.06,const double& sample_decrease_factor=1.5);

        /** \brief build a sweep_surface given the curve c and the orthogonal pattern
            \param c the input curve
            \param pattern the pattern of the orthogonal direction (pattern normal must be (0,0,1))
            \param N_subdiv the number of subdivision between points
            \param is_closed defined is the sweep should be closed or not
            \param N_subdiv is the number of subdivision to close the tube at the 2 extremities
            \param MC_v3d_vector e1 the tangent vector is given to manage the twist
        **/
        static MC_mesh_index_vector sweep_surface(const MC_curve& c,const MC_curve& pattern=MC_curve(),const int& N_subdiv=2,const bool& is_closed=true,const MC_v3d_vector& e1=MC_v3d_vector());

        /** \brief build a unit square with ((N1+2)x(N2+2)) vertices

             do quads indexed by k2+(N1+2)*k1

        ex.
        for(int k1=0;k1<N+2;++k1)
        {
            for(int k2=0;k2<N+2;++k2)
            {
                double u1=double(k1)/double(N-1+2);
                double u2=double(k2)/double(N-1+2);
                mesh_3d.point_set()(k2+k1*(N+2))=MC_v3d(u1,u2,f(u1,u2));
            }
        }

        **/
        static MC_mesh_index_vector build_square(const int& N_1=0,const int& N_2=0);
        /** \brief build a square with given orientation, center and width */
        static MC_mesh_index_vector build_square(const int& N_1,const int& N_2,const MC_v3d& tangent_1,const MC_v3d& tangent_2,const MC_v3d& center,const double& L1,const double& L2);

        /** \brief build a visualization of a set of points
            \param MC_v3d_vector the point_set
            \param the radius of the sphere
            \param int the number of subdivision of the ball
          */
        static MC_mesh_index_vector build_ball_point_set(const MC_v3d_vector& v,const double& radius=0.1,const int& N_subdiv=2);

        /** \brief build a wireframe using the edges */
        static MC_mesh_index_vector build_wireframe(const MC_mesh_index_vector& mesh,const double& radius_cylinder,const int& N_circular,const int& N_subdiv_edges=2);

        /** \brief build a set of sphere given the positions */
        static MC_mesh_index_vector build_sphere_point_set(const MC_v3d_vector& position,const double& radius,const unsigned int& N_subdiv);


        //*********************************************//
        //*********************************************//
        // Math Operator
        //*********************************************//
        //*********************************************//

        /** \brief translate the mesh
            \return the new mesh
        */
        friend MC_mesh_index_vector operator+(const MC_mesh_index_vector& vec,const MC_v3d& to_add);
        /** \brief translate the mesh
            \return the new mesh
        */
        friend MC_mesh_index_vector operator-(const MC_mesh_index_vector& vec,const MC_v3d& to_sub);

        /** \brief translate the mesh
            \return the new mesh
        */
        friend MC_mesh_index_vector operator+(const MC_mesh_index_vector& vec,const MC_v3d_vector& to_add);
        /** \brief translate the mesh
            \return the new mesh
        */
        friend MC_mesh_index_vector operator-(const MC_mesh_index_vector& vec,const MC_v3d_vector& to_sub);

        /** \brief homogeneous scale
            \return the new mesh
        */
        friend MC_mesh_index_vector operator*(const MC_mesh_index_vector& vec,const double& to_mult);
        /** \brief homogeneous scale
            \return the new mesh
        */
        friend MC_mesh_index_vector operator*(const double& to_mult,const MC_mesh_index_vector& vec);
        /** \brief divide a double value to the vector
            \return the new vector
        */
        friend MC_mesh_index_vector operator/(const MC_mesh_index_vector& vec,const double& to_subdiv);



        /** \brief internal translation */
        MC_mesh_index_vector& operator+=(const MC_v3d& to_add);
        /** \brief internal translation */
        MC_mesh_index_vector& operator-=(const MC_v3d& to_sub);
        /** \brief internal translation */
        MC_mesh_index_vector& operator+=(const MC_v3d_vector& to_add);
        /** \brief internal translation */
        MC_mesh_index_vector& operator-=(const MC_v3d_vector& to_sub);
        /** \brief internal scale */
        MC_mesh_index_vector& operator*=(const double& to_mult);
        /** \brief internal scale */
        MC_mesh_index_vector& operator/=(const double& to_subdiv);


        /** \brief apply a matrix transformation to the point set */
        friend MC_mesh_index_vector operator*(const MC_matrix& M,const MC_mesh_index_vector& mesh);
        /** \brief internal matrix multiplication to the point set */
        MC_mesh_index_vector& operator*=(const MC_matrix& M);


        /** \brief return the per vertex normal associated to each vertex smoothed using the 1-ring */
        MC_v3d_vector normal_vertex() const;
        /** \brief return the per vertex normal associated to each vertex smoothed using the 1-ring */
        static MC_v3d_vector normal_vertex(const MC_v3d_vector& point_set,const MC_connectivity_index& connectivity);
        /** \brief return the per vertex normal associated to each vertex smoothed using the 1-ring (for contiguous vector in memory) */
        static MC_double_vector normal_vertex(const MC_double_vector& point_set,const MC_connectivity_index& connectivity);

        /** \brief return the per polygon normal */
        MC_v3d_vector normal_polygon() const;

        //*********************************************//
        //*********************************************//
        // Boundary with geometry
        //*********************************************//
        //*********************************************//

        /** \brief get the boundary curves
            \return the vector of MC_curve of the geometry
            \return the index of the curve in the connectivity of the mesh
        */
        std::pair <std::vector <MC_curve>,MC_int_vector_vector> boundary_curve() const;



        //*********************************************//
        //*********************************************//
        // Add / delete operation with geometry
        //*********************************************//
        //*********************************************//

        /** \brief Fast add one polygons in a unique sens (need a std::set to speed up the find process)

            Add the connectivity and the vertex if it does not exists yet.

            Return a vector containing the n. of connectivity in first.

            In second: is_new the vector 1:if added / 0: if not
        */
        std::pair <MC_int_vector,MC_int_vector> add_unique_polygon(const MC_polygon& polygon,std::map <MC_v3d,int,MC_v3d_less>* map_vertices);


        /** \brief Add a set of polygons in a unique sens (build a std::set to speed up the find process)
            \return the new mesh
            \return a vector containing the n. of connectivity
            \return is_new the vector 1:if added / 0: if not
            \sa build_from_polygon_soup
        */
        std::pair<MC_mesh_index_vector,std::pair<MC_int_vector_vector,MC_int_vector_vector> > added_polygon_soup(const std::vector<MC_polygon> poly_soup) const;

        /** \brief Build a set of polygons in a unique sens (build a std::set to speed up the find process)
            \return the new mesh
            \return a vector containing the n. of connectivity
            \return is_new the vector 1:if added / 0: if not
            \sa added_polygon_soup


            to map to an other curve:
                    std::pair<MC_mesh_index_vector,std::pair<MC_int_vector_vector,MC_int_vector_vector> > mm=MC_mesh_index_vector_draw::build_from_polygon_soup(mesh_every_steps[k].first.get_polygon());

        MC_v3d_vector map_3d;
        for(int k2=0;k2<mm.second.second.size();++k2)
        {
            for(int k3=0;k3<mm.second.second[k2].size();++k3)
            {
                if(mm.second.second[k2][k3]==1)//added vertex
                {
                    map_3d.set(mm.second.first[k2][k3],mesh_every_steps[k].second.get_polygon(k2)[k3]);
                }
            }
        }
        */
        static std::pair<MC_mesh_index_vector,std::pair<MC_int_vector_vector,MC_int_vector_vector> > build_from_polygon_soup(const std::vector<MC_polygon> poly_soup);


        /** \brief get the index of polygons within/outside a given sphere
            \return MC_int_vector: index of polygons inside the sphere
            \return MC_int_vector: index of polygons outside the sphere
        */
        std::pair <MC_int_vector,MC_int_vector> polygons_inside_sphere(const MC_v3d& center,const double& radius) const;

        //*********************************************//
        //*********************************************//
        // Operation with geometry (intersection, ...)
        //*********************************************//
        //*********************************************//

          /** \brief Get the intersection between the mesh and a plane
              \return the curve that intersect the mesh. This curve might be empty if no intersection. It might also be many curves in cases of non convex mesh, or non connected parts

              The plane is defined by its normal n, and a point x0, so its equation is <n,x-x0>=0
          */
        std::vector <MC_curve> plane_intersection(const MC_v3d& n,const MC_v3d& x0) const;

        /** \brief get the intersection of the Mesh and the half space defined by the oriented plane <n,x-x0>=0

            3 Possibilies for type:

            0 - unchanged Mesh

            1 - partly cutted polygon

           2 - no more Mesh at all
        */
        MC_mesh_index_vector half_space_intersection(const MC_v3d& n,const MC_v3d& x0,int *type=0) const;



        /** \brief Return the intersection between Segment and the Mesh
            \return the set of intersection point (MC_v3d_vector)
            \return the corresponding polygon index (MC_int_vector)
            \return the local coordinates of the intersection on the line

            note that the value are ordered in the local coordinate order
        */
        std::pair <MC_v3d_vector,std::pair <MC_int_vector,MC_double_vector> > segment_intersection(const MC_segment& s) const;

        /** \brief helper function for doing picking
            \param a set of meshes and a segment
            \return the index of the selected mesh (MC_int_vector)
            \return the intersection position (MC_v3d_vector)
            \return the index of the intersected polygon in the selected mesh (MC_int_vector)
            \return the relative position of the intersection on the segment (MC_double_vector)
            \sa std::pair <MC_v3d_vector,std::pair <MC_int_vector,MC_double_vector> > segment_intersection(const MC_segment& s) const;

            ordonate the intersection along the increasing relative position
        */
        static std::pair <MC_int_vector,std::pair <MC_v3d_vector,std::pair <MC_int_vector,MC_double_vector> > >
               segment_intersection(const std::vector <MC_mesh_index_vector>& v_mesh,const MC_segment& s);

        /** \brief get the 2D intersection curve between an infinite line and a 2D mesh */
        static MC_curve intersection_curve(const MC_mesh_index_vector& mesh,const MC_segment& c);


        /** \brief map a set of points given by there barycentric coordinates onto an other mesh */
        static MC_v3d_vector map(const MC_mesh_index_vector& mesh_to_map,const MC_double_vector_vector& barycentric_coordinates,const MC_int_vector& polygon_index);

        /** \brief map a set of 3D points from a mesh to an other one using barycentric mapping
            \return MC_v3d_vector the new positions
            \return MC_int_vector the new belonging polygons
            \return MC_double_vector_vector the barycentric coordinates
        */
        static std::pair<MC_v3d_vector,std::pair<MC_int_vector,MC_double_vector_vector> > map(const MC_mesh_index_vector& mesh_to_map,const MC_mesh_index_vector& original_map,const MC_v3d_vector& points_to_map);

        /** \brief project the vector of position onto the surface
            \param MC_mesh_index_vector: the mesh to project onto
            \param MC_v3d_vector: the vertices positions
            \param MC_int_vector: the polygon index where the vertices should be projected to speed up (the polygon must be in the one ring of the closest one)
        */
        static MC_v3d_vector project_to_surface(const MC_mesh_index_vector& mesh_to_project,const MC_v3d_vector& vertex_to_project,const MC_int_vector& triangle_close_to_vertex);
        /** \brief project the position onto the surface
            \param MC_mesh_index_vector: the mesh to project onto
            \param MC_v3d_vector: the vertex
            \param MC_int_vector: the polygon index where the vertex should be projected to speed up (the polygon must be in the one ring of the closest one)
        */
        static MC_v3d_vector project_to_surface(const MC_mesh_index_vector& mesh_to_project,const MC_v3d& vertex_to_project,const int& polygon_close_to_vertex);

        /** \brief get the barycentric coordinate of vertices given their corresponding polygons
            \warning the vertices must lie in their corresponding polygon

            the belonging polygon might be given by MC_mesh_cgal::closest_point(X).second;
        */
        static MC_double_vector_vector barycentric_coordinates(const MC_mesh_index_vector& mesh,const MC_v3d_vector& vertices,const MC_int_vector& belonging_polygon);



        /** \brief delete a set of polygon given by their id
            \return MC_mesh_index_vector the resulting mesh
            \return MC_mesh_index_vector the mesh of the deleted polygon
            \return MC_int_vector the correspondance of the new vertices of the resulting mesh to the old vertices index
            \return MC_int_vector the correspondance of the new vertices of the deleted mesh to the old vertices index
        */
        std::pair <std::pair<MC_mesh_index_vector,MC_mesh_index_vector>,std::pair<MC_int_vector,MC_int_vector> > delete_polygon(const std::set<int> index_to_delete) const;

        /** \brief delete a set of vertex given by their id
            \return MC_mesh_index_vector the mesh without the polygons linking the selected vertices
            \return MC_mesh_index_vector the mesh with the polygons linking the selected vertices
            \return MC_int_vector the correspondance of the index of the vertices of the undeleted part to the original one
            \return MC_int_vector the correspondance of the index of the vertices of the deleted part to the original one

           Add only polygon that are not touching the selected vertices
        */
        std::pair<std::pair<MC_mesh_index_vector,MC_mesh_index_vector>,std::pair<MC_int_vector,MC_int_vector> > delete_vertex(const std::set<int> index_to_delete) const;

        /** \brief delete the polygon touching the border a set of vertex given by their id
            \return MC_mesh_index_vector the mesh without the polygons touching the boundary.
            \return MC_mesh_index_vector the mesh with the polygons touching the boundary.
            \return MC_int_vector the correspondance of the index of the vertices of the undeleted part to the original one
            \return MC_int_vector the correspondance of the index of the vertices of the deleted part to the original one

           Add only polygon that are not touching the selected vertices
        */
        std::pair<std::pair<MC_mesh_index_vector,MC_mesh_index_vector>,std::pair<MC_int_vector,MC_int_vector> > delete_boundary_polygon() const;

        /** \brief paste a surface of mesh onto an other
            \return the new parameterized mesh
            \return a vector of mesh containing : The original mesh with the middle hole, the patch to paste without the ouside vertices, the filled gap between the two surfaces using delaunay
        */
        static std::pair<MC_mesh_index_vector,std::vector<MC_mesh_index_vector> > paste_parameterization(const MC_mesh_index_vector& original_mesh,const MC_mesh_index_vector& patch_to_paste);
        ///slower but more robust
        static std::pair<MC_mesh_index_vector,std::vector<MC_mesh_index_vector> > paste_parameterization_2(const MC_mesh_index_vector& original_mesh,const MC_mesh_index_vector& patch_to_paste);

        //*********************************************//
        //*********************************************//
        // Volume/Area/...
        //*********************************************//
        //*********************************************//

        /** \brief compute the average edge length */
        double average_edge_length() const;

        /** \brief compute the volume of the mesh */
        double volume() const;
        /** \brief compute the gradient of the volume of the mesh */
        MC_v3d_vector volume_gradient() const;

        /** \brief the total area of the mesh */
        double area() const;
        /** \brief compute the gradient of the area of the mesh */
        MC_v3d_vector area_gradient() const;


        /** \brief compute barycentric centroid based weighted by polygon area
            \param MC_int_vector selected_polygon: the polgons that should be taken into account (default empty: means every polygons)
        */
        MC_v3d centroid_polygon(const MC_int_vector& selected_polygon=MC_int_vector()) const;

        /** \brief compute inertia matrix */
        MC_matrix inertia() const;

        //*********************************************//
        //*********************************************//
        // Deformation
        //*********************************************//
        //*********************************************//

        /** \brief Laplacian deformation using mean value
            \param MC_mesh_index_vector& the input mesh (position+connectivity)
            \param std::pair<MC_int_vector,MC_v3d_vector> the input constraints (index and value)
        */
        static MC_mesh_index_vector laplacian_deformation(const MC_mesh_index_vector& input_mesh,const std::pair<MC_int_vector,MC_v3d_vector>& constraints);

        /** \brief get the polygon transformation between two meshes
            \return the vector of matrix transformation such that m1=R*m0
        */
        static std::vector<MC_matrix> polygon_transformation(const MC_mesh_index_vector& m0,const MC_mesh_index_vector& m1);

        /** \brief perform an as rigid as possible deformation between a reference and a given mesh (polar decomposition per triangle)

            ex. MC_mesh_index_vector::as_rigid_as_possible(m0,m1,std::make_pair(0,m1.point_set()(0)));
        */
        static MC_mesh_index_vector as_rigid_as_possible(const MC_mesh_index_vector& m0,const MC_mesh_index_vector& m1,const std::pair<MC_int_vector,MC_v3d_vector>& constraints);

        /** \brief compute the stretch between two vector of polygons */
        static std::pair<std::vector<MC_matrix>,std::pair<MC_v3d_vector,MC_v3d_vector> > stretch(const MC_mesh_index_vector& m0,const MC_mesh_index_vector& m1);

        static std::pair<std::vector<MC_matrix>,std::pair<MC_v3d_vector,MC_v3d_vector> > UGLY_stretch_filter(const MC_mesh_index_vector& m0,const MC_mesh_index_vector& m1,const int& N_step=3);


        /** \brief Laplacian smoothing
            \param double lambda: the smoothing factor in [0,1] (default=0.5)
            \param int steps: the number of smoothing steps (default=1)
            \param bool is_boundary_preserved: does the deformation preserves the boundary vertices or not (default=true)
            \return the smoothed mesh
        */
        MC_mesh_index_vector laplacian_smoothing(const double& lambda=0.5,const int& steps=1,const bool& is_boundary_preserving=true) const;

        //*********************************************//
        //*********************************************//
        // Angle
        //*********************************************//
        //*********************************************//

        /** \brief compute angular error between two meshes */
        static std::pair<double,MC_double_vector> angular_error(const MC_mesh_index_vector& mesh_0,const MC_mesh_index_vector& mesh_1);

        //*********************************************//
        //*********************************************//
        // Loader
        //*********************************************//
        //*********************************************//

        /** \brief load a mesh from a file if the extension is recognized */
        static MC_mesh_index_vector load_mesh_file(const std::string& filename);



        //*********************************************//
        //*********************************************//
        // Convert geometry to texture coordinate
        //*********************************************//
        //*********************************************//

        /** \brief convert to a planar texture map in projecting onto the xy-plane and normalized in [0,1] */
        static MC_mesh_index_vector texture_convert_planar_xy(const MC_mesh_index_vector& mesh);

        /** \brief convert to a planar texture map and ensure the same connectivity between texture and mesh_ref */
        static MC_mesh_index_vector texture_convert_planar_xy_same_connectivity(const MC_mesh_index_vector& mesh,const MC_mesh_index_vector& mesh_ref);


    protected:

        /** \brief internal storage of the connectivity*/
        MC_connectivity_index connectivity_mesh;
        /** \brief internal storage of the point set*/
        MC_v3d_vector point_set_mesh;

    };
}

#endif
