
/*
**    Mesh Converter
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef _MC_MESH_INDEX_VECTOR_DRAW_HPP_
#define _MC_MESH_INDEX_VECTOR_DRAW_HPP_

#include <MC_mesh_index_vector.hpp>

#include <MC_v3d_vector.hpp>
#include <MC_double_vector.hpp>

namespace mesh_conv
{


    /** \brief A class of indexed mesh with color, normal and texture ready to be (slowly) drawable but with easily access */
    class MC_mesh_index_vector_draw : public MC_mesh_index_vector
    {
    public:

        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \brief empty constructor */
        MC_mesh_index_vector_draw();
        /** \brief direct point_set+connectivity constructor */
        MC_mesh_index_vector_draw(const MC_v3d_vector& input_point_set,const MC_connectivity_index& input_connectivity);
        /** \brief direct constructor from a MC_mesh_index_vector (vertex+connectivity) */
        MC_mesh_index_vector_draw(const MC_mesh_index_vector& mesh);
        /** \brief copy constructor  */
        MC_mesh_index_vector_draw(const MC_mesh_index_vector_draw& mesh);

        //*********************************************//
        //*********************************************//
        // Get
        //*********************************************//
        //*********************************************//

        /** \brief get the normal */
        MC_v3d_vector& normal();
        /** \brief get the normal */
        const MC_v3d_vector& normal() const;
        /** \brief get the color */
        MC_v3d_vector& color();
        /** \brief get the color */
        const MC_v3d_vector& color() const;
        /** \brief get the texture */
        MC_v3d_vector& texture();
        /** \brief get the texture */
        const MC_v3d_vector& texture() const;
        /** \brief get the texture connectivity */
        const MC_connectivity_index& texture_connectivity() const;
        /** \brief get the texture connectivity */
        MC_connectivity_index& texture_connectivity();
        /** \brief get the alpha */
        MC_double_vector& alpha();
        /** \brief get the alpha */
        const MC_double_vector& alpha() const;


        //*********************************************//
        //*********************************************//
        // COLOR MANAGER
        //*********************************************//
        //*********************************************//

        /** \brief fill the whole mesh with the same color */
        MC_mesh_index_vector_draw& fill_color(const MC_v3d& color);
        /** \brief fill the whole mesh with a per vertex color */
        MC_mesh_index_vector_draw& fill_color(const MC_v3d_vector& color);

        //*********************************************//
        //*********************************************//
        // Texture
        //*********************************************//
        //*********************************************//

        /** \brief set the texture */
        MC_mesh_index_vector_draw& set_texture(const std::pair<MC_v3d_vector,MC_connectivity_index>& _texture);
        /** \brief set the texture with a texture_mesh input */
        MC_mesh_index_vector_draw& set_texture(const MC_mesh_index_vector& texture_mesh);

        //*********************************************//
        //*********************************************//
        // Special shape
        //*********************************************//
        //*********************************************//

        /** \brief create local axis visualization
            \return the set of meshes (ux,uy,uz,xyz_ball)

            exemple of final visualization:

            std::vector <MC_mesh_index_vector> v_axis=MC_mesh_index_vector::build_local_basis();
            for(unsigned int k=0;k<v_axis.size();k++)
              openGL_drawer_basis.add_draw_list(v_axis[k]);

            glPushMatrix();
            glLoadIdentity();
            glTranslated(0.4,0.4,-1.25);
            glMultMatrixd(navigator.get_quaternion_matrix_1().get_pointer());
            openGL_drawer_basis.draw_stored_list();
            glPopMatrix();
        */
        static std::vector <MC_mesh_index_vector_draw> build_local_basis();

        //*********************************************//
        //*********************************************//
        // Manipulation
        //*********************************************//
        //*********************************************//

        /** \brief concatenation of the point set/normal/color/texture and the connectivity
            \return a new mesh in directly adding the vertices

            the connectivity is automatically incremented to avoid conflict
        */
        friend MC_mesh_index_vector_draw operator<<(const MC_mesh_index_vector_draw& vec0,const MC_mesh_index_vector_draw& vec1);

        /** \brief internal concatenation of the pointset/normal/color/texture and connectivity
            \return a new mesh in directly adding the vertices

            the connectivity is automatically incremented to avoid conflict
        */
        MC_mesh_index_vector_draw& concatenation(const MC_mesh_index_vector_draw& vec1);


        //*********************************************//
        //*********************************************//
        // Math
        //*********************************************//
        //*********************************************//

        /** \brief point_set multiplication by a matrix */
        friend MC_mesh_index_vector_draw operator*(const MC_matrix& M,const MC_mesh_index_vector_draw& mesh);
        /** \brief point_set multiplication by a matrix */
        friend MC_mesh_index_vector_draw operator*(const MC_mesh_index_vector_draw& mesh,const MC_matrix& M);

        /** \brief homogeneous scale */
        friend MC_mesh_index_vector_draw operator*(const double& s,const MC_mesh_index_vector_draw& mesh);
        /** \brief homogeneous scale */
        friend MC_mesh_index_vector_draw operator*(const MC_mesh_index_vector_draw& mesh,const double& s);

        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator+(const MC_mesh_index_vector_draw& vec,const MC_v3d& to_add);
        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator-(const MC_mesh_index_vector_draw& vec,const MC_v3d& to_sub);
        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator+(const MC_mesh_index_vector_draw& vec,const MC_v3d_vector& to_add);
        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator-(const MC_mesh_index_vector_draw& vec,const MC_v3d_vector& to_sub);

        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator+(const MC_v3d& to_add,const MC_mesh_index_vector_draw& vec);
        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator-(const MC_v3d& to_sub,const MC_mesh_index_vector_draw& vec);
        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator+(const MC_v3d_vector& to_add,const MC_mesh_index_vector_draw& vec);
        /** \brief translate the mesh */
        friend MC_mesh_index_vector_draw operator-(const MC_v3d_vector& to_sub,const MC_mesh_index_vector_draw& vec);

        /** \brief divide a double value to the vector */
        friend MC_mesh_index_vector_draw operator/(const MC_mesh_index_vector_draw& vec,const double& to_subdiv);

        /** \brief internal translation */
        MC_mesh_index_vector_draw& operator+=(const MC_v3d& to_add);
        /** \brief internal translation */
        MC_mesh_index_vector_draw& operator-=(const MC_v3d& to_sub);
        /** \brief internal translation */
        MC_mesh_index_vector_draw& operator+=(const MC_v3d_vector& to_add);
        /** \brief internal translation */
        MC_mesh_index_vector_draw& operator-=(const MC_v3d_vector& to_sub);
        /** \brief internal scale */
        MC_mesh_index_vector_draw& operator*=(const double& to_mult);
        /** \brief internal scale */
        MC_mesh_index_vector_draw& operator/=(const double& to_subdiv);



        //*********************************************//
        //*********************************************//
        // Normal (fast)
        //*********************************************//
        //*********************************************//

        /** \brief update per vertex normal to the current mesh */
        MC_mesh_index_vector_draw& normal_vertex_update();


        //*********************************************//
        //*********************************************//
        // Loader
        //*********************************************//
        //*********************************************//

        /** \brief load a mesh file if its extension is recognized

            Compared to MC_mesh_index_vector::load_mesh_file(), the MC_mesh_index_vector_draw try to load texture coordinates for some extensions
        */
        static MC_mesh_index_vector_draw load_mesh_file(const std::string& filename);


        //*********************************************//
        //*********************************************//
        // Subdivision
        //*********************************************//
        //*********************************************//

        /** \brief subdivide mixed mid edge*/
        MC_mesh_index_vector_draw subdivide_mixed_mid_edge(const std::set<int>& subdivided_polygon) const;

    private:

        /** \brief the vector of normal */
        MC_v3d_vector normal_mesh;
        /** \brief the vector of color */
        MC_v3d_vector color_mesh;
        /** \brief the vector of texture (uv-map, only x-y coordinates are read) */
        MC_v3d_vector texture_mesh;
        /** \brief the connectivity of the texture */
        MC_connectivity_index texture_connectivity_mesh;
        /** \brief alpha vector */
        MC_double_vector alpha_mesh;


    };

}

#endif
