/*
**    Mesh Converter
**    Copyright (C) 2008 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MC_OPENGL_DRAWER_HPP_
#define MC_OPENGL_DRAWER_HPP_

#include <stdlib.h>
#include <iostream>

#include <GL/glut.h>
#include <MC_curve.hpp>

namespace mesh_conv
{

    class MC_mesh_index_vector;
    class MC_mesh_index_vector_draw;
    class MC_mesh_fast_draw;
    class MC_segment;
    class MC_v3d;
    class MC_polygon;

    namespace object
    {
        class MC_bounding_box_oriented;
    }

    namespace mesh
    {
        namespace cgal
        {
            class MC_mesh_cgal;
        }
    }

    /// Class of drawer in Opengl
    /** Interface between data class and OpenGL stuff */
    class MC_opengl_drawer
    {
    public:


        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//


        /** \brief empty constructor*/
        MC_opengl_drawer();



        //*********************************************//
        //*********************************************//
        // Static drawer
        //*********************************************//
        //*********************************************//

        /** \brief draw a MC_mesh_index_vector */
        static const MC_mesh_index_vector& draw(const MC_mesh_index_vector& mesh);
        /** \brief draw a MC_mesh_index_vector with a per polygon color*/
        static void draw_per_polygon_color(const MC_mesh_index_vector& mesh,const MC_v3d_vector& color_per_polygon);
        /** \brief draw a MC_mesh_index_vector_draw */
        static const MC_mesh_index_vector_draw& draw(const MC_mesh_index_vector_draw& mesh);
        /** \brief draw a MC_mesh_fast_draw */
        static const MC_mesh_fast_draw& draw(const MC_mesh_fast_draw& mesh);
        /** \brief draw a MC_segment */
        static const MC_segment& draw(const MC_segment& segment);
        /** \brief draw a vector of MC_segment */
        static const std::vector<MC_segment>& draw(const std::vector<MC_segment>& segments);
        /** \brief draw a MC_curve */
        static const MC_curve& draw(const MC_curve& curve);
        /** \brief draw a vector of MC_curve */
        static const std::vector <MC_curve>& draw(const std::vector <MC_curve>& curves);
        /** \brief draw a MC_polygon */
        static const MC_polygon& draw(const MC_polygon& polygon);
        /** \brief draw an oriented bounding box */
        static void draw(const object::MC_bounding_box_oriented& bb);

        /** \brief draw a MC_mesh_cgal (the mesh sould be made of triangles only)*/
        static MC_opengl_drawer draw_triangular(const mesh::cgal::MC_mesh_cgal& mesh);

        /** \brief draw directly a sphere of given position and radius */
        static void draw_sphere(const MC_v3d& position,const double& radius=0.5,const int& N_subdiv=2);
        /** \brief draw directly a sphere of given position and radius

            compute only one occurence of the sphere and plot the same at different place using glMatrix translation
        */
        static void draw_sphere(const MC_v3d_vector& position,const double& radius=0.5,const int& N_subdiv=2);

        /** \brief draw sampled points */
        static void draw_points(const MC_v3d_vector& position);

        /** \brief draw a grid of a mesh */
        static const MC_mesh_index_vector& draw_grid(const MC_mesh_index_vector& mesh);
        /** \brief draw a grid of a fast draw mesh */
        static const MC_mesh_fast_draw& draw_grid(const MC_mesh_fast_draw& mesh);

        /** \brief draw the per-polygon normals */
        static const MC_mesh_index_vector& draw_normal_per_polygon(const MC_mesh_index_vector& mesh,const double& L=0.1);

        /** \brief draw a vector of drawable structure */
        template <typename T>
                static const MC_opengl_drawer draw(const std::vector<T>& to_draw)
        {
            for(int k=0,N=to_draw.size();k<N;++k)
                MC_opengl_drawer::draw(to_draw[k]);
            return MC_opengl_drawer();
        }

        //*********************************************//
        //*********************************************//
        // Send geometry to GPU
        //*********************************************//
        //*********************************************//

        /** \brief send a geometry to the GPU (use VBO or draw_list depending on the graphics card)
            \return the openGL identifier to call for drawing */
        static GLuint send_gpu(const MC_mesh_fast_draw& mesh);

        /** \brief send a geometry grid to the GPU
            \return the openGL identifier to call for drawing */
        static GLuint send_gpu_grid(const MC_mesh_index_vector& mesh);

    private:


    };
}

#endif
