/*
**    Mesh Converter
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MC_POLYGON_HPP_
#define MC_POLYGON_HPP_

#include <MC_v3d.hpp>
#include <MC_v3d_vector.hpp>
#include <MC_int_vector.hpp>

#include <stdlib.h>
#include <iostream>



namespace mesh_conv
{
    class MC_segment;
    class MC_mesh_index_vector;

    /// Container class for a polygon
    /** Internal structure is a vector of MC_v3d */
    class MC_polygon : public MC_v3d_vector
    {
    public:

        //************************************************//
        //************************************************//
        //Constructor
        //************************************************//
        //************************************************//

        /** \brief empty constructor */
        MC_polygon();
        /** \brief constructor from a triangle */
        MC_polygon(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2);
        /** \brief constructor from a quad */
        MC_polygon(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2,const MC_v3d& v3);
         /** \brief constructor from a pentagone */
        MC_polygon(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2,const MC_v3d& v3,const MC_v3d& v4);
        /** \brief copy constructor */
        MC_polygon(const MC_v3d_vector& vec);
        /** \brief direct constructor from a vector of MC_v3d */
        MC_polygon(const std::vector <MC_v3d>& vec);
        /** \brief copy constructor */
        MC_polygon(const MC_polygon& p);


        //************************************************//
        //************************************************//
        //Informations
        //************************************************//
        //************************************************//

        /** \brief check if the polygon is planar.
            \return false if not planar, true if planar.
        */
        bool is_planar() const;

        /** \brief get the normal of the polygon (assume it is almost planar).
            \return the normal as a R3 vector.
        */
        MC_v3d normal() const;

        /** \brief get the barycenter of the polygon */
        MC_v3d barycenter() const;

        /** \brief check if a vertex (in plane) is inside the polygon
            \return true if inside, false outside */
        bool is_inside(const MC_v3d& _x) const;
        /** \brief check if a vertex (in plane) is outside the polygon
            \return true if outside, false inside */
        bool is_outside(const MC_v3d& _x) const;


        /** \brief get shortest position on the polygon between the polygon and the point
            \return the closest point

            type the type of point as pointer (default=null)

            type = 0: interior point in the polygon

            type = 1: edge point

            type = 2: vertex point

            A vertex is inside the polygon (in the plane) if it lies on one of its triangulation
        */
        MC_v3d closest_point(const MC_v3d& x,int *type=0) const;


        /** \brief get the generalized barycentric coordinates
            \warning x must be inside the polygon

            see [Meyer, Lee, Barr, Desbrun: Generalized Barycentric Coordinates on Irregular Polygons]
        */
        MC_double_vector barycentric_coordinates(const MC_v3d& x) const;


        /** \brief return the type of position of a vertex with respect to the polygon
            \return (0,-1) inside the polygon, (1,k_edge) on an edge, (2,k_vertex) on a vertex, (3,-1) not in the polygon
        */
        MC_int_vector position(const MC_v3d& x) const;





        //************************************************//
        //************************************************//
        //Geometry transform (intersection, ...)
        //************************************************//
        //************************************************//


        /** \brief Intersection with the plane of equation <n,x-x0>=0
            \param n: the normal
            \param x0: a point of the plane
            \param type: the type of intersection (defaut null)
            \param type_edge: the detailed interesection type by edge (defaut null)

            The plane is defined by its normal n, and a point x0, so its equation is <n,x-x0>=0

            4 possibilities:

            _ no intersections (type 0) => return void vector

            _ plane intersect on two distinct points (type 1) => return vector of size 2

            _ plane intersect only on a vertex (type 2) => return vector of size 1

            _ plane intersect along an edge of the Polygon (type 3 > return the two extreme vertices) => return a vector of size 2

         */
        MC_v3d_vector plane_intersection(const MC_v3d& n,const MC_v3d& x0,int *type=0,MC_int_vector *type_edge=0) const;

        /** \brief get the intersection of the polygon and the half space defined by the oriented plane <n,x-x0>=0
            \param n the normal of the intersecting-plane
            \param x0 a point on the plane
            \param type for the type of intersection (default 0)
            \return a polygon resulting of the intersection

            3 Possibilies for type:

            0 - unchanged polygon

            1 - partly cutted polygon

            2 - no more polygon at all
        */
        MC_polygon half_space_intersection(const MC_v3d& n,const MC_v3d& x0,int *type=0) const;

        /** \brief segment_intersection
            \param s the segment to intersect
            \param type the type of intersection (default 0)
            \return the point of intersection if it exist (else defaut value)

            _ no intersections (type 0)

            _ triangle intersection on one interior point (type 1)

            _ triangle intersection on a vertex of the Segment (type 2)

            _ triangle is partly or totally inside the segment (type 3)  */
        MC_v3d segment_intersection(const MC_segment& s,int *type=0) const;


        /** \brief project a point in the given direction (coplanar to the polygon)
            \warning, the direction must be coplanar to the polygon
            \return the position of the projected point, its type and its relative coordinates
            \return V_3D: position in 3D space
            \return 1st int: type (0=vertex, 1=edge, -1=no_intersection)
            \return 2nd int: vertex/edge number in the polygon (-1=no_intersection)
            \return double : relative position (0 for edge, interpolation for edge)
        */
        std::pair <MC_v3d,std::pair<int,std::pair<int,double> > > projected_direction(const MC_v3d& x0,const MC_v3d& d,const MC_int_vector& forbidden_edge=MC_int_vector(-1)) const;



        /** \brief check if the polygon is degenerated (same vertices) */
        bool is_degenerated() const;
        /** \brief return a polygon not degenerated if possible
            \return the new undegenerated polygon and a set of indicators:

            1.does the polygon has been changed

            2.does the new polygon is valid
        */
        std::pair<MC_polygon,std::pair<bool,bool> > undegenerated() const;



        /** \brief rotate the polygon into a plane defined by its normal (if polygon is planar)
            \return the new polygon
            \return the matrix of rotation
            \note the matrix is not unique (twist around the normal)
        */
        std::pair <MC_polygon,MC_matrix> rotate_to_plane(const MC_v3d& normal) const;


        //************************************************//
        //************************************************//
        //Subdivision
        //************************************************//
        //************************************************//.

        /** \brief split the given polygon by linking consecutiv mid edges but with constraint on the given edge to not add extra mid edge
            \return std::vector <MC_polygon> the vector of new polygons
            \return MC_mesh_index the mesh corresponding to the complete polygons
            \return MC_int_vector the index of the added vertices
            \return MC_int_vector the original edge corresponding to the newly created indices

            Creates only triangles excepted for the last one which links every mid_points
            Works well for triangles
        */
        std::pair<std::vector <MC_polygon>,std::pair<MC_mesh_index_vector,std::pair<MC_int_vector,MC_int_vector> > >  subdivide_mid_edge(const MC_int_vector& constraint_edges=MC_int_vector()) const;



        /** \brief split the given polygon by linking barycenter to the consecutiv mid edges but with constraint on the given edge to not add extra mid edge
            \return std::vector <MC_polygon> the vector of new polygons
            \return MC_mesh_index the mesh corresponding to the complete polygons
            \return MC_int_vector the index of the added vertices (barycenter added anyway at position size())
            \return MC_int_vector the original edge corresponding to the newly created indices

            Creates only triangles excepted for the last one which links every mid_points

            Works well for quads
        */
        std::pair<std::vector <MC_polygon>,std::pair<MC_mesh_index_vector,std::pair<MC_int_vector,MC_int_vector> > > subdivide_barycenter_mid_edge(const MC_int_vector& constraint_edges=MC_int_vector()) const;

        /** \brief subdivision of vertices adapted to polygon type

            Use subdivide_barycenter_mid_edge for size()>3 and subdivide_mid_edge for size()==3
        */
        std::pair<std::vector <MC_polygon>,std::pair<MC_mesh_index_vector,std::pair<MC_int_vector,MC_int_vector> > > subdivide_mixed_mid_edge(const MC_int_vector& constraint_edges=MC_int_vector()) const;


        //************************************************//
        //************************************************//
        //Get
        //************************************************//
        //************************************************//

        /** \brief get the segment of the given edge */
         MC_segment segment(const int& edge_number) const;


    private:

        /** \brief split the given polygon by linking consecutiv edges
            \return eturn the set of new polygons

            Creates only triangles excepted for the last one which links every mid_points

            Works well for triangles
        */
        std::vector <MC_polygon> subdivide_mid_edge_no_constraint() const;
        /** \brief split the given polygon by linking barycenter to the middle edges
            \return Return the set of new polygons

            Creates only four vertices polygons

            Works well for quads
        */
        std::vector <MC_polygon> subdivide_barycenter_mid_edge_no_constraint() const;


    };


    class MC_polygon_less
    {
    public:
        bool operator()(const MC_polygon& p0,const MC_polygon& p1) const
        {
            MC_v3d_less L;
            if(p0.size()<p1.size())
                return true;
            else if(p0.size()>p1.size())
                return false;
            else //equal size compare by values
            {
                int N=p0.size();
                for(int k=0;k<N;k++)
                {
                    if(L(p0[k],p1[k])==true)
                        return true;
                    else if(L(p1[k],p0[k])==true)
                        return false;
                }
                //complete equality
                return false;
            }
        }
    };

}




#endif
