/*
**    Mesh Converter
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef MC_MC_QUATERNION_HPP
#define MC_MC_QUATERNION_HPP

#include <iostream>
#include <cmath>

#include <MC_v4d.hpp>
#include <vector>




namespace mesh_conv
{

    class MC_v3d;
    class MC_v3d_vector;
    class MC_double_vector;

    /// Class of quaternion
    /** internally based on v4d (axis_x,axis_y,axis_z,angle) */

    class MC_quaternion : public MC_v4d
    {
    public:

        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \brief zero contructor (set unitary axis (0,0,1) and zero angle ) */
        MC_quaternion();
        /** \brief constructor from axis and angle */
        MC_quaternion(const MC_v3d& axis,const double& angle);
        /** \brief constructor from base class: v4d */
        MC_quaternion(const MC_v4d& v);
        /** \brief copy constructor */
        MC_quaternion(const MC_quaternion& q);
        /** \brief direct constructor from coefficients */
        MC_quaternion(const double& x,const double& y,const double& z,const double& w);



        //*********************************************//
        //*********************************************//
        // Math Operator
        //*********************************************//
        //*********************************************//

        /** \brief compute the rotation matrix associated to the current quaternion
            \return a 3x3 rotation matrix
        */
        MC_matrix matrix() const;

        /** \brief get the conjugate of the current quaterion */
        MC_quaternion conjugated() const;

        /** \brief multiplication between quaternion */
        friend MC_quaternion operator*(const MC_quaternion& q0,const MC_quaternion& q1);
        /** \brief internal multiplication between quaternion */
        MC_quaternion& operator*=(const MC_quaternion& q1);



        /** \brief power of a quaternion with a double
            \return exp(t*log_q(q)) */
        MC_quaternion pow(const double& t) const;

        /** \brief invert the quaternion
            \return Q^-1 = (Q*)/||Q|| */
        MC_quaternion inverted() const;


        /** \brief log map

             the log is going from S3 -> R3 */
        MC_v3d log() const;

        /** \brief exp map

            the exp is going from R3 -> S3 */
        static MC_quaternion exp(const MC_v3d& v);

        /** \brief Spherical Linear Interpolation SLERP

            Use (sin((1-t) theta )q0 +sin(t theta)q1)/sin(theta)
            But same results with (q1 q0^{-1})^{t} q0
            (q1*q0.invert()).pow_q(t)*q0) */
        static MC_quaternion slerp(const MC_quaternion& q0,const MC_quaternion& q1,const double& t);


        /** Generalized approximate spherical interpolation for rotation

            Compute sum(qi)/norm(sum(qi)) */
        static MC_quaternion quat_interp(const std::vector <MC_quaternion>& quaternion_vec,const MC_double_vector& weights);
        /** \brief Generalized approximate spherical interpolation for rotation and linear one for translation
            \return a 4x4 matrix
        */
        static MC_matrix quat_interp(const std::vector <MC_matrix>& matrix_vec,const MC_double_vector& weights);
        /** \brief Generalized approximate spherical interpolation for rotation and linear one for translation */
        static MC_matrix quat_interp(const std::vector <MC_quaternion>& quaternion_vec,const MC_double_vector& weights,const MC_v3d_vector& translation_vec);


        //*********************************************//
        //*********************************************//
        // GET
        //*********************************************//
        //*********************************************//

        /** \brief get the unit axis corresponding to the quaternion */
        MC_v3d axis() const;
        /** \brief get the angle corresponding to the quaternion */
        double angle() const;

        //*********************************************//
        //*********************************************//
        // OUTPUT
        //*********************************************//
        //*********************************************//


    private:


    };



}

#endif // MC_v4d_HPP
