
/*
**    Mesh Converter
**    Copyright (C) 2009  Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef _MC_SEGMENT_HPP_
#define _MC_SEGMENT_HPP_


#include <stdlib.h>
#include <iostream>

#include <MC_v3d.hpp>
#include <vector>

namespace mesh_conv
{

    class MC_double_vector;
    class MC_int_pair;


    /** \brief A class to manipulate segments given by two points of R3 */
    class MC_segment
    {
        public:


        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \brief classical constructor */
        MC_segment();
        /** \brief direct constructor */
        MC_segment(const MC_v3d& start_point,const MC_v3d& end_point);
        /** \brief Copy constructor */
        MC_segment(const MC_segment& s);

        /** destructor */
        ~MC_segment();


        //*********************************************//
        //*********************************************//
        // Manipulation
        //*********************************************//
        //*********************************************//

        /** \brief get the vector (x0,x1) / ||(x0,x1)|| */
        MC_v3d unit_vector() const;
        /** \brief get the vector x1-x0 */
        MC_v3d vector() const;

        /** \brief inverse the segment (first <-> last) */
        MC_segment fliped() const;

        // ***************************************************** //
        // ***************************************************** //
        // Build
        // ***************************************************** //
        // ***************************************************** //

        /** \brief build a vector of segment from a point_set and a vector of index */
        static std::vector<MC_segment> build_segment_vector(const MC_v3d_vector& vertices,const std::vector<MC_int_pair>& constraints);



        //*********************************************//
        //*********************************************//
        // Information
        //*********************************************//
        //*********************************************//

        /** \brief return the length of the segment */
        double length() const;

        /** \brief return the closest point of x on the segment */
        MC_v3d closest_point(const MC_v3d& x) const;

        /** \brief return the distance of point x to the closest point on the segment */
        double distance_to_point(const MC_v3d& x) const;

        /** \brief return the segment of minimal length between two segments

            There is only 2 posibilities

            1. The closest point links the extremities

            2. The closest point is the perpendicular to both segments

           (The first point of the returned segment belongs to the current segment)
        */
        MC_segment closest_to_segment(const MC_segment& s) const;


        /** \brief get the minimal distance between the two segments

            based on the function closest_point_in_segment */
        double distance_to_segment(const MC_segment& s) const;

        /** \brief Intersection with the plane of equation <n,x-x0>=0
            \param n the normal of the plane
            \param x0 a point within the plane
            \param type a pointer to return the type if needed (default NULL)

            The plane is defined by its normal n, and a point x0, so its equation is <n,x-x0>=0

            4 possibilities: _ no intersections (type 0)

                             _ plane intersect on one interior points (type 1)

                             _ plane intersect on a vertex (type 2)

                             _ plane contains the segment (type 3)

             Solving the system <n,x-x0>=0 & A+t*u=x leads to t=<n,x0-A>/<n,u>
        */
        MC_v3d plane_intersection(const MC_v3d& n,const MC_v3d& x0,int *type=0) const;

        /** \brief check if a point belongs to the segment
            \return true if inside the segment, false if not
            \param x the position of the point
            \param the type of belonging (1: inside the segment, -1: aligned but outside of it, -2: not aligned with the direction) */
        bool is_belonging(const MC_v3d& x,int *type=0) const;

        /** \brief check if a point belongs to the direction of the line

            does not need to be inside the segment */
        bool is_aligned(const MC_v3d& x) const;


        /** \brief if a point is aligned with the direction, get its parameterization t such that (x0,x)=t (x0,x1)
            \param x the current position
            \param is_aligned return 1 if the point belongs to the direction, 0 if not (default NULL) */
        double relative_position(const MC_v3d& x,bool *is_aligned=0) const;


        /** \brief assert the index is either 0 or 1
            \return nothing, but stop the programm if it goes outside bounds
        */
        void assert_bounds(const int& u) const;


        //************************************************//
        //************************************************//
        //Get value
        //************************************************//
        //************************************************//

        /** \brief direct access operator */
        MC_v3d  operator()(const int& index) const;
        /** \brief direct access operator */
        MC_v3d& operator()(const int& index);
        /** \brief direct access operator */
        MC_v3d  operator[](const int& index) const;
        /** \brief direct access operator */
        MC_v3d& operator[](const int& index);

        /** \brief get linear interpolation at time t (parameterization bw [0,1])
            \param t the time of the interpolation
            \return x0*(1-t)+x1*t
        */
        MC_v3d value(const double& t) const;

        /** \brief get linear interpolation at time t (parameterization bw [0,1])
            \param t the vector time of the interpolation
            \return x0*(1-t)+x1*t in a vector form
        */
        MC_v3d_vector value(const MC_double_vector& t) const;

        /** \brief subsampling every d_L
            \param d_L the approximate sampling length
            \return the set of linear samples and the corresponding t value
            \sa linear_sampling(int N)

            does not constraint exactly d_L, but make sure it is an exact number of subdivision
        */
        std::pair<MC_v3d_vector,MC_double_vector> linear_sampling_intervals(const double& d_L) const;

        /** \brief subsampling in N samples
            \param N the number of sections
            \return the set of linear samples and its corresponding parameters
        */
        std::pair<MC_v3d_vector,MC_double_vector> linear_sampling(const int& N) const;



        //************************************************//
        //************************************************//
        //Math operator
        //************************************************//
        //************************************************//


        /** \brief translate the segment */
        friend MC_segment operator+(const MC_segment& s,const MC_v3d& x);
        /** \brief translate the segment */
        friend MC_segment operator+(const MC_v3d& x,const MC_segment& s);
        /** \brief scale the segment */
        friend MC_segment operator*(const MC_segment& s,const double& a);
        /** \brief scale the segment */
        friend MC_segment operator*(const double& a,const MC_segment& s);


        //************************************************//
        //************************************************//
        //OUTPUT
        //************************************************//
        //************************************************//

        /** Ostream operator */
        friend std::ostream& operator << (std::ostream& stream, const MC_segment& s);

    private:

        /** \brief The internal data structure is two MC_v3d */
        MC_v3d x0,x1;

    };
}

#endif
