/*
**    Mesh Converter
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef MC_TRIANGLE_HPP_
#define MC_TRIANGLE_HPP_

#include <MC_v3d.hpp>
#include <MC_v3d_vector.hpp>
#include <MC_polygon.hpp>
#include <MC_matrix.hpp>

#include <stdlib.h>
#include <iostream>



namespace mesh_conv
{

    class MC_segment;

    /// Container class for a triangle
    /** Internal structure is a vector of MC_v3d */
    class MC_triangle : public MC_v3d_vector
    {
    public:

        //************************************************//
        //************************************************//
        //Constructor
        //************************************************//
        //************************************************//

        /** \brief empty constructor */
        MC_triangle();
        /** \brief constructor from a triangle */
        MC_triangle(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2);
        /** \brief copy constructor */
        MC_triangle(const MC_v3d_vector& vec);
        /** \brief direct constructor from a vector of MC_v3d */
        MC_triangle(const std::vector <MC_v3d>& vec);
        /** \brief copy constructor */
        MC_triangle(const MC_triangle& p);

        /** \brief constructor from polygon (is 3 edge) */
        MC_triangle(const MC_polygon& p);

        /** \brief triangulate polygon (do fan starting at p[0])*/
        static std::vector<MC_triangle> triangulate(const MC_polygon& p);

        //************************************************//
        //************************************************//
        //Informations
        //************************************************//
        //************************************************//

        /** \brief get the normal of the triangle.
            \return the normal as a R3 vector.
        */
        MC_v3d normal() const;

        /** \brief get the barycenter of the polygon */
        MC_v3d barycenter() const;

        //************************************************//
        //************************************************//
        // Informations
        //************************************************//
        //************************************************//

        /** \brief get shortest position on the triangle between the triangle and the point
            \return the closest point

            type the type of point as pointer (default=null)

            type = 0: interior point in the polygon

            type = 1: edge point

            type = 2: vertex point

            A vertex is inside the polygon (in the plane) if it lies on one of its triangulation
        */
        MC_v3d closest_point(const MC_v3d& x,int *type=0) const;


        /** \brief check if a vertex (in plane) is inside the triangle by calculating its barycentric coordinates
            \return true if inside, false outside */
        bool is_inside(const MC_v3d& _x) const;

        /** \brief get barycentric coordinates for vertex in this plane
            \return double_vector (alpha,beta,gamma) if x is in the plane,
            \return true/false if x is in the plane */
        std::pair<MC_double_vector,bool> barycentric_coordinates(const MC_v3d& x) const;

        /** \brief get the area of the triangle */
        double area() const;

        /** \brief get the inertia matrix (3x3) of a triangle

           see http://en.wikipedia.org/wiki/Inertia_tensor_of_triangle
           J=tr(C)*I-C, where C=int(x*x^t)
        */
        MC_matrix inertia() const;


        //************************************************//
        //************************************************//
        //Geometry transform (intersection, ...)
        //************************************************//
        //************************************************//

        /** \brief segment_intersection
            \param s the segment to intersect
            \param type the type of intersection (default 0)
            \return the point of intersection if it exist (else defaut value)

            _ no intersections (type 0)

            _ triangle intersection on one interior point (type 1)

            _ triangle intersection on a vertex of the Segment (type 2)

            _ triangle is partly or totally inside the segment (type 3)  */
        MC_v3d segment_intersection(const MC_segment& s,int *type=0) const;


        /** \brief PN triangle
          

            [Curved PN triangles, A. Vlachos] */
        static MC_v3d pn_triangle(const MC_v3d_vector& X,const MC_v3d_vector& N,const MC_double_vector& bar_coord) ;

        //************************************************//
        //************************************************//
        // Get
        //************************************************//
        //************************************************//

        /** \brief get the segment of the given edge */
        MC_segment segment(const int& edge_number) const;


    private:

    };

}

#endif
