/*
**    Mesh Converter
**    Copyright (C) 2009 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef MC_V3D_HPP
#define MC_V3D_HPP

#include <iostream>
#include <cmath>





namespace mesh_conv
{
    class MC_double_vector;
    class MC_v3d_vector;
    class MC_matrix;
    class MC_quaternion;
    class MC_segment;

/// Basic class for 3D vector
/** Internal structure is a static table of double[3] */

class MC_v3d
{
public:

    //*********************************************//
    //*********************************************//
    // CONSTRUCTORS
    //*********************************************//
    //*********************************************//

    /** \brief empty constructor of zero vector */
    MC_v3d();
    /** \brief direct constructor from x,y,z */
    MC_v3d(const double& x,const double& y,const double& z);
    /** \brief copy constructor */
    MC_v3d(const MC_v3d& _v);
    /** \brief constructor from pointer to double
        \warning the pointer must have a size >=2
    */
    MC_v3d(const double* _v);
    /** \brief constructor from a string "(x,y,z)"

      exemple of degenerated cases "(5,;x 4,g)"->MC_v3d(5,4,0)
    */
    MC_v3d(const std::string& s);

    /** \brief constructor from a matrix only if size=(3,1); (4,1) or (1,3); (1,4) */
    MC_v3d(const MC_matrix& M);

    /** \brief constructor from a MC_v3d_vector if size==1 */
    MC_v3d(const MC_v3d_vector& vec);

    /** \brief constructor from a MC_double_vector if size==3 */
    MC_v3d(const MC_double_vector& vec);

    /** \brief destructor */
    ~MC_v3d();


    // ********************************************* //
    // ********************************************* //
    //  Math Operator
    // ********************************************* //
    // ********************************************* //

    /** \brief set the value to 0 */
    void set_zero();


    //*********************************************//
    //*********************************************//
    // Equality
    //*********************************************//
    //*********************************************//

    /** \brief operator= from a matrix only if size=(3,1); (4,1) or (1,3); (1,4)*/
    MC_v3d& operator=(const MC_matrix& M);

    //*********************************************//
    //*********************************************//
    // Math Operator
    //*********************************************//
    //*********************************************//

    /** \brief change the opposite sign of the current vector (unary negation)
        \return a new vector with opposite sign
     */
    MC_v3d operator-() const;

    /** \brief add a double value to the vector
        \return the new vector
     */
    friend MC_v3d operator+(const MC_v3d& vec,const double& to_add);
    /** \brief substract a double value to the vector
        \return the new vector
     */
    friend MC_v3d operator-(const MC_v3d& vec,const double& to_sub);
    /** \brief multiply a double value to the vector
        \return the new vector
     */
    friend MC_v3d operator*(const MC_v3d& vec,const double& to_mult);
    /** \brief multiply a double value to the vector
        \return the new vector
     */
    friend MC_v3d operator*(const double& to_mult,const MC_v3d& vec);
    /** \brief divide a double value to the vector
        \return the new vector
     */
    friend MC_v3d operator/(const MC_v3d& vec,const double& to_subdiv);


    /** \brief add a v3d to the vector
        \return the new vector
     */
    friend MC_v3d operator+(const MC_v3d& vec,const MC_v3d& to_add);
    /** \brief substract a v3d to the vector
        \return the new vector
     */
    friend MC_v3d operator-(const MC_v3d& vec,const MC_v3d& to_sub);


    /** \brief internal addition with a double */
    MC_v3d& operator+=(const double& to_add);
    /** \brief internal substraction with a double */
    MC_v3d& operator-=(const double& to_sub);
    /** \brief internal multiplication with a double*/
    MC_v3d& operator*=(const double& to_mult);
    /** \brief internal divide with a double */
    MC_v3d& operator/=(const double& to_subdiv);

    /** \brief internal addition */
    MC_v3d& operator+=(const MC_v3d& to_add);
    /** \brief internal substraction */
    MC_v3d& operator-=(const MC_v3d& to_sub);

    /** \brief internal scaling the vector
        \return the new vector
    */
    MC_v3d& scale(const double& sx,const double& sy,const double& sz);
    /** \brief internal scaling the vector
        \return the new vector
    */
    MC_v3d& scale(const MC_v3d& scaling);
    /** \brief internal homogeneous scaling the vector
        \return the new vector
    */
    MC_v3d& scale(const double& scaling);

    /** \brief scaling the vector
        \return the new vector
    */
    MC_v3d scaled(const MC_v3d& scaling) const;
    /** \brief homogeneous scaling the vector
        \return the new vector
    */
    MC_v3d scaled(const double& scaling) const;


    /** \brief apply a pointwise mask to the vector (do not change it)
        \return the masked vector
    */
    MC_v3d mask(const double& sx,const double& sy,const double& sz) const;
    /** \brief apply a pointwise mask the vector
        \return the masked vector
    */
    MC_v3d mask(const MC_v3d& scaling) const;
    /** \brief apply a pointwise mask to the vector
        \return the masked vector
    */
    MC_v3d scale(const double& scaling) const;

    /** \brief dot product between two vectors */
    double dot(const MC_v3d& vec) const;
    /** \brief cross product */
    MC_v3d cross(const MC_v3d& vec) const;
    /** \brief norm of the vector */
    double norm() const;
    /** \brief get the normalized vector.
        \return a vector of norm 1 (return (0,0,1) if the vector is 0)
    */
    MC_v3d normalized() const;


    /** \brief internal application of a matrix to the vector of MC_v3d (does the product x'=M*x) */
    MC_v3d& operator*=(const MC_matrix& M);

    /** \brief apply the quaternion rotation to a MC_v3d
        \return q v q^b */
    friend MC_v3d operator*(const MC_quaternion& q,const MC_v3d& vec);
    /** \brief internal quaternion rotation application to a MC_v3d
        \return q v q^b */
    MC_v3d& operator*=(const MC_quaternion& q);


    /** \brief boolean equality using MC_v3d_less */
    friend bool operator==(const MC_v3d& a1,const MC_v3d& a2);
    /** \brief boolean non-equality using MC_v3d_less */
    friend bool operator!=(const MC_v3d& a1,const MC_v3d& a2);


    /** \brief get the area spanning two vectors
        \return ||v0 x v1||/2
    */
    static double area(const MC_v3d& v0,const MC_v3d& v1);

    /** \brief cotan between 2 vectors
        \return (u0.u1) / ||u0 x u1||
   */
    static double cotan(const MC_v3d& u0,const MC_v3d& u1);

    /** \brief angle between two vectors
        \return acos ((u0.u1)/(||u0|| ||u1||))
      */
    static double angle(const MC_v3d& u0,const MC_v3d& u1);

    /** \brief project a vector onto plane
        \param MC_v3d the normal of the plane
        \return u=u-(u.n)n/||n||^2
    */
    MC_v3d project_on_plane(const MC_v3d& normal) const;

    /** \brief project a point onto line directed by the vector u
        \param Vector u such that line=span(u)
    */
    MC_v3d project_on_line(const MC_v3d& dir) const;
    /** \brief project a point onto line directed by the segment
        \param Segment seg such that line=seg[0]+span(seg[1]-seg[0])
    */
    MC_v3d project_on_line(const MC_segment& seg) const;

    //*********************************************//
    //*********************************************//
    // Operator with gradient
    //*********************************************//
    //*********************************************//

    std::pair <MC_v3d,MC_v3d_vector> normalized_with_gradient() const;


    //*********************************************//
    //*********************************************//
    // GET
    //*********************************************//
    //*********************************************//

    /** \brief get operator (k_dim must be 0,1, or 2) */
    const double& operator()(const int& k_dim) const;
    /** \brief get operator (k_dim must be 0,1, or 2) */
    double& operator()(const int& k_dim);
    /** \brief get operator (k_dim must be 0,1, or 2) */
    const double& operator[](const int& k_dim) const;
    /** \brief get operator (k_dim must be 0,1, or 2) */
    double& operator[](const int& k_dim);

    /** \brief return the pointer to the internal data */
    const double* pointer() const;
    /** \brief return the pointer to the internal data
        \warning unprotected
    */
    double* pointer_unprotected();

    //*********************************************//
    //*********************************************//
    // OUTPUT
    //*********************************************//
    //*********************************************//

    /** \brief write the vector under the form (x,y,z) */
    friend std::ostream& operator<<(std::ostream& stream,const MC_v3d& v);

    /** \brief write the v3d to string (use operator <<) */
    std::string to_string() const;

    /** \brief read the vector from a istream under the form (x,y,z) */
    friend bool operator>>(std::istream& stream,MC_v3d& v);

private:

    /** \brief internal struct */
    double v[3];
};


/// Comparator class for set and map
class MC_v3d_less
{
 public:
  bool operator ()(const MC_v3d& v0,const MC_v3d& v1) const
  {

    double epsilon=1e-5;
    if( std::fabs(v0[0]-v1[0])<epsilon && std::fabs(v0[1]-v1[1])<epsilon && std::fabs(v0[2]-v1[2])<epsilon )
      return false;


    if(v0[0]+epsilon<v1[0])
      return true;
    else if ( !(v1[0]+epsilon<v0[0]) && v0[1]+epsilon<v1[1] )
      return true;
    else if ( !(v1[0]+epsilon<v0[0]) && !(v1[1]+epsilon<v0[1]) && v0[2]+epsilon<v1[2] )
      return true;


    return false;
  }
};

}

#endif // MC_V3D_HPP
