/*
**    Mesh Converter
**    Copyright (C) 2008 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/



#ifndef MC_V4D_HPP
#define MC_V4D_HPP

#include <iostream>
#include <cmath>





namespace mesh_conv
{

    class MC_v3d;
    class MC_matrix;

/// Basic class for 4D vector
/** Internal structure is a static table of double[4] */

class MC_v4d
{
public:

    //*********************************************//
    //*********************************************//
    // CONSTRUCTORS
    //*********************************************//
    //*********************************************//

    /** \brief empty constructor of zero vector */
    MC_v4d();
    /** \brief direct constructor from x,y,z,w */
    MC_v4d(const double& x,const double& y,const double& z,const double& w);
    /** \brief copy constructor */
    MC_v4d(const MC_v4d& _v);
    /** \brief constructor from pointer to double
        \warning the pointer must have a size >=3
    */
    MC_v4d(const double* _v);
    /** \brief constructor from a string "(x,y,z,w)"

      exemple of degenerated cases "(5,;x 4,g,8)"->MC_v4d(5,4,0,8)
    */
    MC_v4d(const std::string& s);

    /** \brief constructor from a matrix only if size=(4,1) or (1,4) */
    MC_v4d(const MC_matrix& M);

    /** \brief constructor from a MC_v3d (add 1 at the end) */
    MC_v4d(const MC_v3d& vec);

    /** \brief destructor */
    ~MC_v4d();




    //*********************************************//
    //*********************************************//
    // Math Operator
    //*********************************************//
    //*********************************************//

    /** \brief change the opposite sign of the current vector (unary negation)
        \return a new vector with opposite sign
     */
    MC_v4d operator-() const;

    /** \brief add a double value to the vector
        \return the new vector
     */
    friend MC_v4d operator+(const MC_v4d& vec,const double& to_add);
    /** \brief substract a double value to the vector
        \return the new vector
     */
    friend MC_v4d operator-(const MC_v4d& vec,const double& to_sub);
    /** \brief multiply a double value to the vector
        \return the new vector
     */
    friend MC_v4d operator*(const MC_v4d& vec,const double& to_mult);
    /** \brief multiply a double value to the vector
        \return the new vector
     */
    friend MC_v4d operator*(const double& to_mult,const MC_v4d& vec);
    /** \brief divide a double value to the vector
        \return the new vector
     */
    friend MC_v4d operator/(const MC_v4d& vec,const double& to_subdiv);


    /** \brief add a v4d to the vector
        \return the new vector
     */
    friend MC_v4d operator+(const MC_v4d& vec,const MC_v4d& to_add);
    /** \brief substract a v4d to the vector
        \return the new vector
     */
    friend MC_v4d operator-(const MC_v4d& vec,const MC_v4d& to_sub);


    /** \brief internal addition with a double */
    MC_v4d& operator+=(const double& to_add);
    /** \brief internal substraction with a double */
    MC_v4d& operator-=(const double& to_sub);
    /** \brief internal multiplication with a double*/
    MC_v4d& operator*=(const double& to_mult);
    /** \brief internal divide with a double */
    MC_v4d& operator/=(const double& to_subdiv);

    /** \brief internal addition */
    MC_v4d& operator+=(const MC_v4d& to_add);
    /** \brief internal substraction */
    MC_v4d& operator-=(const MC_v4d& to_sub);

    /** \brief internal scaling the vector
        \return the new vector
    */
    MC_v4d& scale(const double& sx,const double& sy,const double& sz,const double& sw);
    /** \brief internal scaling the vector
        \return the new vector
    */
    MC_v4d& scale(const MC_v4d& scaling);
    /** \brief internal homogeneous scaling the vector
        \return the new vector
    */
    MC_v4d& scale(const double& scaling);

    /** \brief apply a pointwise mask to the vector (do not change it)
        \return the masked vector
    */
    MC_v4d mask(const double& sx,const double& sy,const double& sz,const double& sw) const;
    /** \brief apply a pointwise mask the vector
        \return the masked vector
    */
    MC_v4d mask(const MC_v4d& scaling) const;
    /** \brief apply a pointwise mask to the vector
        \return the masked vector
    */
    MC_v4d scale(const double& scaling) const;

    /** \brief dot product between two vectors */
    double dot(const MC_v4d& vec) const;
    /** \brief norm of the vector */
    double norm() const;
    /** \brief get the normalized vector.
        \return a vector of norm 1 (return (0,0,1,0) if the vector is 0)
    */
    MC_v4d normalized() const;


    /** \brief internal application of a matrix to the vector of MC_v4d (does the product x'=M*x) */
    MC_v4d& operator*=(const MC_matrix& M);


    /** \brief set only the 3d part without changing the 4th component */
    MC_v4d& set_v3d(const MC_v3d& vec);


    //*********************************************//
    //*********************************************//
    // GET
    //*********************************************//
    //*********************************************//

    /** \brief get operator (k_dim must be 0,1,2 or 3) */
    const double& operator()(const int& k_dim) const;
    /** \brief get operator (k_dim must be 0,1,2 or 3) */
    double& operator()(const int& k_dim);
    /** \brief get operator (k_dim must be 0,1,2 or 3) */
    const double& operator[](const int& k_dim) const;
    /** \brief get operator (k_dim must be 0,1,2 or 3) */
    double& operator[](const int& k_dim);

    /** \brief get the pointer on the vector
        \warning use with care
    */
    const double* pointer() const;

    //*********************************************//
    //*********************************************//
    // OUTPUT
    //*********************************************//
    //*********************************************//

    /** \brief write the vector under the form (x,y,z,w) */
    friend std::ostream& operator<<(std::ostream& stream,const MC_v4d& v);

    /** \brief read the vector from a istream under the form (x,y,z,w) */
    friend MC_v4d& operator>>(std::istream& stream,MC_v4d& v);

private:

    /** \brief internal struct */
    double v[4];
};


/// Comparator class for set and map
class MC_v4d_less
{
 public:
  bool operator ()(const MC_v4d& v0,const MC_v4d& v1) const
  {

    double epsilon=0.000001;
    if( fabs(v0[0]-v1[0])<epsilon && fabs(v0[1]-v1[1])<epsilon && fabs(v0[2]-v1[2])<epsilon )
      return false;


    if(v0[0]<v1[0])
      return true;
    else if ( !(v1[0]<v0[0]) && v0[1]<v1[1] )
      return true;
    else if ( !(v1[0]<v0[0]) && !(v1[1]<v0[1]) && v0[2]<v1[2] )
      return true;
    else if ( !(v1[0]<v0[0]) && !(v1[1]<v0[1]) && v0[2]<v1[2] && v0[3]<v1[3] )
      return true;

    return false;
  }
};

}

#endif // MC_v4d_HPP
